# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'logic_settings_window.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_LogicSettingsWindow(object):
    def setupUi(self, LogicSettingsWindow):
        if not LogicSettingsWindow.objectName():
            LogicSettingsWindow.setObjectName(u"LogicSettingsWindow")
        LogicSettingsWindow.resize(765, 702)
        self.verticalLayout = QVBoxLayout(LogicSettingsWindow)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 4, 0, 0)
        self.menuBar = QMenuBar(LogicSettingsWindow)
        self.menuBar.setObjectName(u"menuBar")

        self.verticalLayout.addWidget(self.menuBar)

        self.name_layout = QHBoxLayout()
        self.name_layout.setSpacing(6)
        self.name_layout.setObjectName(u"name_layout")
        self.name_label = QLabel(LogicSettingsWindow)
        self.name_label.setObjectName(u"name_label")

        self.name_layout.addWidget(self.name_label)

        self.name_edit = QLineEdit(LogicSettingsWindow)
        self.name_edit.setObjectName(u"name_edit")

        self.name_layout.addWidget(self.name_edit)


        self.verticalLayout.addLayout(self.name_layout)

        self.main_tab_widget = QTabWidget(LogicSettingsWindow)
        self.main_tab_widget.setObjectName(u"main_tab_widget")
        self.logic_tab = QWidget()
        self.logic_tab.setObjectName(u"logic_tab")
        self.logic_tab_layout = QVBoxLayout(self.logic_tab)
        self.logic_tab_layout.setSpacing(6)
        self.logic_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.logic_tab_layout.setObjectName(u"logic_tab_layout")
        self.logic_tab_layout.setContentsMargins(0, 0, 0, 0)
        self.logic_tab_widget = QTabWidget(self.logic_tab)
        self.logic_tab_widget.setObjectName(u"logic_tab_widget")
        self.trick_level_tab = QWidget()
        self.trick_level_tab.setObjectName(u"trick_level_tab")
        self.trick_level_top_layout = QVBoxLayout(self.trick_level_tab)
        self.trick_level_top_layout.setSpacing(6)
        self.trick_level_top_layout.setContentsMargins(11, 11, 11, 11)
        self.trick_level_top_layout.setObjectName(u"trick_level_top_layout")
        self.trick_level_top_layout.setContentsMargins(0, 0, 0, 0)
        self.trick_level_scroll = QScrollArea(self.trick_level_tab)
        self.trick_level_scroll.setObjectName(u"trick_level_scroll")
        self.trick_level_scroll.setFrameShape(QFrame.NoFrame)
        self.trick_level_scroll.setFrameShadow(QFrame.Plain)
        self.trick_level_scroll.setWidgetResizable(True)
        self.trick_level_scroll_contents = QWidget()
        self.trick_level_scroll_contents.setObjectName(u"trick_level_scroll_contents")
        self.trick_level_scroll_contents.setGeometry(QRect(0, 0, 757, 559))
        self.trick_level_layout = QVBoxLayout(self.trick_level_scroll_contents)
        self.trick_level_layout.setSpacing(6)
        self.trick_level_layout.setContentsMargins(11, 11, 11, 11)
        self.trick_level_layout.setObjectName(u"trick_level_layout")
        self.trick_level_layout.setContentsMargins(4, 8, 4, 0)
        self.logic_description_label = QLabel(self.trick_level_scroll_contents)
        self.logic_description_label.setObjectName(u"logic_description_label")
        self.logic_description_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.logic_description_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.logic_description_label)

        self.trick_level_line_1 = QFrame(self.trick_level_scroll_contents)
        self.trick_level_line_1.setObjectName(u"trick_level_line_1")
        self.trick_level_line_1.setFrameShape(QFrame.HLine)
        self.trick_level_line_1.setFrameShadow(QFrame.Sunken)

        self.trick_level_layout.addWidget(self.trick_level_line_1)

        self.trick_level_minimal_logic_check = QCheckBox(self.trick_level_scroll_contents)
        self.trick_level_minimal_logic_check.setObjectName(u"trick_level_minimal_logic_check")

        self.trick_level_layout.addWidget(self.trick_level_minimal_logic_check)

        self.trick_level_minimal_logic_label = QLabel(self.trick_level_scroll_contents)
        self.trick_level_minimal_logic_label.setObjectName(u"trick_level_minimal_logic_label")
        self.trick_level_minimal_logic_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.trick_level_minimal_logic_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.trick_level_minimal_logic_label)

        self.trick_level_line_2 = QFrame(self.trick_level_scroll_contents)
        self.trick_level_line_2.setObjectName(u"trick_level_line_2")
        self.trick_level_line_2.setFrameShape(QFrame.HLine)
        self.trick_level_line_2.setFrameShadow(QFrame.Sunken)

        self.trick_level_layout.addWidget(self.trick_level_line_2)

        self.trick_level_help_label = QLabel(self.trick_level_scroll_contents)
        self.trick_level_help_label.setObjectName(u"trick_level_help_label")
        self.trick_level_help_label.setWordWrap(True)

        self.trick_level_layout.addWidget(self.trick_level_help_label)

        self.trick_level_scroll.setWidget(self.trick_level_scroll_contents)

        self.trick_level_top_layout.addWidget(self.trick_level_scroll)

        self.logic_tab_widget.addTab(self.trick_level_tab, "")
        self.logic_damage_tab = QWidget()
        self.logic_damage_tab.setObjectName(u"logic_damage_tab")
        self.logic_damage_layout = QVBoxLayout(self.logic_damage_tab)
        self.logic_damage_layout.setSpacing(6)
        self.logic_damage_layout.setContentsMargins(11, 11, 11, 11)
        self.logic_damage_layout.setObjectName(u"logic_damage_layout")
        self.logic_damage_layout.setContentsMargins(0, 0, 0, 0)
        self.damage_strictness_group = QGroupBox(self.logic_damage_tab)
        self.damage_strictness_group.setObjectName(u"damage_strictness_group")
        self.damage_strictness_layout = QVBoxLayout(self.damage_strictness_group)
        self.damage_strictness_layout.setSpacing(6)
        self.damage_strictness_layout.setContentsMargins(11, 11, 11, 11)
        self.damage_strictness_layout.setObjectName(u"damage_strictness_layout")
        self.damage_strictness_label = QLabel(self.damage_strictness_group)
        self.damage_strictness_label.setObjectName(u"damage_strictness_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.damage_strictness_label.sizePolicy().hasHeightForWidth())
        self.damage_strictness_label.setSizePolicy(sizePolicy)
        self.damage_strictness_label.setWordWrap(True)

        self.damage_strictness_layout.addWidget(self.damage_strictness_label)

        self.damage_strictness_spacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.damage_strictness_layout.addItem(self.damage_strictness_spacer)

        self.damage_strictness_combo = QComboBox(self.damage_strictness_group)
        self.damage_strictness_combo.addItem("")
        self.damage_strictness_combo.addItem("")
        self.damage_strictness_combo.addItem("")
        self.damage_strictness_combo.setObjectName(u"damage_strictness_combo")

        self.damage_strictness_layout.addWidget(self.damage_strictness_combo)


        self.logic_damage_layout.addWidget(self.damage_strictness_group)

        self.damage_strictness_spacer_2 = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.logic_damage_layout.addItem(self.damage_strictness_spacer_2)

        self.logic_tab_widget.addTab(self.logic_damage_tab, "")
        self.location_pool_tab = QWidget()
        self.location_pool_tab.setObjectName(u"location_pool_tab")
        self.location_pool_layout = QVBoxLayout(self.location_pool_tab)
        self.location_pool_layout.setSpacing(6)
        self.location_pool_layout.setContentsMargins(11, 11, 11, 11)
        self.location_pool_layout.setObjectName(u"location_pool_layout")
        self.location_pool_layout.setContentsMargins(0, 0, 0, 0)
        self.randomization_mode_group = QGroupBox(self.location_pool_tab)
        self.randomization_mode_group.setObjectName(u"randomization_mode_group")
        self.verticalLayout_2 = QVBoxLayout(self.randomization_mode_group)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.randomization_mode_label = QLabel(self.randomization_mode_group)
        self.randomization_mode_label.setObjectName(u"randomization_mode_label")
        self.randomization_mode_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.randomization_mode_label)

        self.randomization_mode_combo = QComboBox(self.randomization_mode_group)
        self.randomization_mode_combo.addItem("")
        self.randomization_mode_combo.addItem("")
        self.randomization_mode_combo.setObjectName(u"randomization_mode_combo")

        self.verticalLayout_2.addWidget(self.randomization_mode_combo)


        self.location_pool_layout.addWidget(self.randomization_mode_group)

        self.excluded_locations_group = QGroupBox(self.location_pool_tab)
        self.excluded_locations_group.setObjectName(u"excluded_locations_group")
        self.verticalLayout_5 = QVBoxLayout(self.excluded_locations_group)
        self.verticalLayout_5.setSpacing(6)
        self.verticalLayout_5.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(2, 3, 2, 0)
        self.excluded_locations_label = QLabel(self.excluded_locations_group)
        self.excluded_locations_label.setObjectName(u"excluded_locations_label")

        self.verticalLayout_5.addWidget(self.excluded_locations_label)

        self.excluded_locations_area = QScrollArea(self.excluded_locations_group)
        self.excluded_locations_area.setObjectName(u"excluded_locations_area")
        self.excluded_locations_area.setWidgetResizable(True)
        self.excluded_locations_area_contents = QWidget()
        self.excluded_locations_area_contents.setObjectName(u"excluded_locations_area_contents")
        self.excluded_locations_area_contents.setGeometry(QRect(0, 0, 745, 356))
        self.excluded_locations_area_layout = QHBoxLayout(self.excluded_locations_area_contents)
        self.excluded_locations_area_layout.setSpacing(6)
        self.excluded_locations_area_layout.setContentsMargins(11, 11, 11, 11)
        self.excluded_locations_area_layout.setObjectName(u"excluded_locations_area_layout")
        self.excluded_locations_area_layout.setContentsMargins(0, 0, 0, 0)
        self.excluded_locations_area.setWidget(self.excluded_locations_area_contents)

        self.verticalLayout_5.addWidget(self.excluded_locations_area)


        self.location_pool_layout.addWidget(self.excluded_locations_group)

        self.logic_tab_widget.addTab(self.location_pool_tab, "")

        self.logic_tab_layout.addWidget(self.logic_tab_widget)

        self.main_tab_widget.addTab(self.logic_tab, "")
        self.patches_tab = QWidget()
        self.patches_tab.setObjectName(u"patches_tab")
        self.patches_tab_layout = QVBoxLayout(self.patches_tab)
        self.patches_tab_layout.setSpacing(6)
        self.patches_tab_layout.setContentsMargins(11, 11, 11, 11)
        self.patches_tab_layout.setObjectName(u"patches_tab_layout")
        self.patches_tab_layout.setContentsMargins(0, 0, 0, 0)
        self.patches_tab_widget = QTabWidget(self.patches_tab)
        self.patches_tab_widget.setObjectName(u"patches_tab_widget")
        self.energy_tab = QWidget()
        self.energy_tab.setObjectName(u"energy_tab")
        self.energy_layout = QVBoxLayout(self.energy_tab)
        self.energy_layout.setSpacing(6)
        self.energy_layout.setContentsMargins(11, 11, 11, 11)
        self.energy_layout.setObjectName(u"energy_layout")
        self.energy_layout.setContentsMargins(0, 0, 0, 0)
        self.energy_tank_box = QGroupBox(self.energy_tab)
        self.energy_tank_box.setObjectName(u"energy_tank_box")
        self.gridLayout_2 = QGridLayout(self.energy_tank_box)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.dangerous_tank_check = QCheckBox(self.energy_tank_box)
        self.dangerous_tank_check.setObjectName(u"dangerous_tank_check")

        self.gridLayout_2.addWidget(self.dangerous_tank_check, 4, 0, 1, 2)

        self.energy_tank_capacity_description = QLabel(self.energy_tank_box)
        self.energy_tank_capacity_description.setObjectName(u"energy_tank_capacity_description")
        self.energy_tank_capacity_description.setWordWrap(True)

        self.gridLayout_2.addWidget(self.energy_tank_capacity_description, 0, 0, 1, 3)

        self.energy_tank_capacity_spin_box = QSpinBox(self.energy_tank_box)
        self.energy_tank_capacity_spin_box.setObjectName(u"energy_tank_capacity_spin_box")
        self.energy_tank_capacity_spin_box.setMinimum(2)
        self.energy_tank_capacity_spin_box.setMaximum(1000)

        self.gridLayout_2.addWidget(self.energy_tank_capacity_spin_box, 2, 2, 1, 1)

        self.energy_tank_capacity_spacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.gridLayout_2.addItem(self.energy_tank_capacity_spacer, 1, 1, 1, 1)

        self.line = QFrame(self.energy_tank_box)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.gridLayout_2.addWidget(self.line, 3, 0, 1, 3)

        self.energy_tank_capacity_label = QLabel(self.energy_tank_box)
        self.energy_tank_capacity_label.setObjectName(u"energy_tank_capacity_label")

        self.gridLayout_2.addWidget(self.energy_tank_capacity_label, 2, 0, 1, 1)


        self.energy_layout.addWidget(self.energy_tank_box)

        self.safe_zone_box = QGroupBox(self.energy_tab)
        self.safe_zone_box.setObjectName(u"safe_zone_box")
        self.gridLayout = QGridLayout(self.safe_zone_box)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.safe_zone_regen_spin = QDoubleSpinBox(self.safe_zone_box)
        self.safe_zone_regen_spin.setObjectName(u"safe_zone_regen_spin")
        self.safe_zone_regen_spin.setMaximum(100.000000000000000)
        self.safe_zone_regen_spin.setSingleStep(0.100000000000000)
        self.safe_zone_regen_spin.setValue(1.000000000000000)

        self.gridLayout.addWidget(self.safe_zone_regen_spin, 2, 1, 1, 1)

        self.safe_zone_regen_label = QLabel(self.safe_zone_box)
        self.safe_zone_regen_label.setObjectName(u"safe_zone_regen_label")

        self.gridLayout.addWidget(self.safe_zone_regen_label, 2, 0, 1, 1)

        self.safe_zone_logic_heal_check = QCheckBox(self.safe_zone_box)
        self.safe_zone_logic_heal_check.setObjectName(u"safe_zone_logic_heal_check")
        self.safe_zone_logic_heal_check.setEnabled(False)
        self.safe_zone_logic_heal_check.setChecked(True)

        self.gridLayout.addWidget(self.safe_zone_logic_heal_check, 1, 0, 1, 2)

        self.safe_zone_description = QLabel(self.safe_zone_box)
        self.safe_zone_description.setObjectName(u"safe_zone_description")

        self.gridLayout.addWidget(self.safe_zone_description, 0, 0, 1, 2)


        self.energy_layout.addWidget(self.safe_zone_box)

        self.dark_aether_box = QGroupBox(self.energy_tab)
        self.dark_aether_box.setObjectName(u"dark_aether_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.dark_aether_box.sizePolicy().hasHeightForWidth())
        self.dark_aether_box.setSizePolicy(sizePolicy1)
        self.dark_aether_layout_2 = QGridLayout(self.dark_aether_box)
        self.dark_aether_layout_2.setSpacing(6)
        self.dark_aether_layout_2.setContentsMargins(11, 11, 11, 11)
        self.dark_aether_layout_2.setObjectName(u"dark_aether_layout_2")
        self.varia_suit_spin_box = QDoubleSpinBox(self.dark_aether_box)
        self.varia_suit_spin_box.setObjectName(u"varia_suit_spin_box")
        self.varia_suit_spin_box.setMaximum(60.000000000000000)
        self.varia_suit_spin_box.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
        self.varia_suit_spin_box.setValue(6.000000000000000)

        self.dark_aether_layout_2.addWidget(self.varia_suit_spin_box, 2, 1, 1, 1)

        self.varia_suit_label = QLabel(self.dark_aether_box)
        self.varia_suit_label.setObjectName(u"varia_suit_label")

        self.dark_aether_layout_2.addWidget(self.varia_suit_label, 2, 0, 1, 1)

        self.dark_suit_label = QLabel(self.dark_aether_box)
        self.dark_suit_label.setObjectName(u"dark_suit_label")

        self.dark_aether_layout_2.addWidget(self.dark_suit_label, 3, 0, 1, 1)

        self.dark_suit_spin_box = QDoubleSpinBox(self.dark_aether_box)
        self.dark_suit_spin_box.setObjectName(u"dark_suit_spin_box")
        self.dark_suit_spin_box.setMaximum(60.000000000000000)
        self.dark_suit_spin_box.setStepType(QAbstractSpinBox.AdaptiveDecimalStepType)
        self.dark_suit_spin_box.setValue(1.200000000000000)

        self.dark_aether_layout_2.addWidget(self.dark_suit_spin_box, 3, 1, 1, 1)

        self.dark_aether_label = QLabel(self.dark_aether_box)
        self.dark_aether_label.setObjectName(u"dark_aether_label")
        sizePolicy.setHeightForWidth(self.dark_aether_label.sizePolicy().hasHeightForWidth())
        self.dark_aether_label.setSizePolicy(sizePolicy)
        self.dark_aether_label.setWordWrap(False)

        self.dark_aether_layout_2.addWidget(self.dark_aether_label, 0, 0, 1, 2)

        self.dark_aether_box_spacer = QSpacerItem(20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed)

        self.dark_aether_layout_2.addItem(self.dark_aether_box_spacer, 1, 0, 1, 2)


        self.energy_layout.addWidget(self.dark_aether_box)

        self.energy_tank_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.energy_layout.addItem(self.energy_tank_spacer)

        self.patches_tab_widget.addTab(self.energy_tab, "")
        self.elevator_tab = QWidget()
        self.elevator_tab.setObjectName(u"elevator_tab")
        self.elevator_layout = QVBoxLayout(self.elevator_tab)
        self.elevator_layout.setSpacing(6)
        self.elevator_layout.setContentsMargins(11, 11, 11, 11)
        self.elevator_layout.setObjectName(u"elevator_layout")
        self.elevator_layout.setContentsMargins(4, 8, 4, 0)
        self.elevators_description_label = QLabel(self.elevator_tab)
        self.elevators_description_label.setObjectName(u"elevators_description_label")
        self.elevators_description_label.setWordWrap(True)

        self.elevator_layout.addWidget(self.elevators_description_label)

        self.elevators_combo = QComboBox(self.elevator_tab)
        self.elevators_combo.addItem("")
        self.elevators_combo.addItem("")
        self.elevators_combo.addItem("")
        self.elevators_combo.addItem("")
        self.elevators_combo.addItem("")
        self.elevators_combo.setObjectName(u"elevators_combo")

        self.elevator_layout.addWidget(self.elevators_combo)

        self.patches_tab_widget.addTab(self.elevator_tab, "")
        self.starting_area_tab = QWidget()
        self.starting_area_tab.setObjectName(u"starting_area_tab")
        self.starting_area_layout = QVBoxLayout(self.starting_area_tab)
        self.starting_area_layout.setSpacing(6)
        self.starting_area_layout.setContentsMargins(11, 11, 11, 11)
        self.starting_area_layout.setObjectName(u"starting_area_layout")
        self.starting_area_layout.setContentsMargins(4, 8, 4, 0)
        self.startingarea_description = QLabel(self.starting_area_tab)
        self.startingarea_description.setObjectName(u"startingarea_description")
        self.startingarea_description.setWordWrap(True)

        self.starting_area_layout.addWidget(self.startingarea_description)

        self.starting_area_quick_fill_layout = QHBoxLayout()
        self.starting_area_quick_fill_layout.setSpacing(6)
        self.starting_area_quick_fill_layout.setObjectName(u"starting_area_quick_fill_layout")
        self.starting_area_quick_fill_label = QLabel(self.starting_area_tab)
        self.starting_area_quick_fill_label.setObjectName(u"starting_area_quick_fill_label")

        self.starting_area_quick_fill_layout.addWidget(self.starting_area_quick_fill_label)

        self.starting_area_quick_fill_ship = QPushButton(self.starting_area_tab)
        self.starting_area_quick_fill_ship.setObjectName(u"starting_area_quick_fill_ship")

        self.starting_area_quick_fill_layout.addWidget(self.starting_area_quick_fill_ship)

        self.starting_area_quick_fill_save_station = QPushButton(self.starting_area_tab)
        self.starting_area_quick_fill_save_station.setObjectName(u"starting_area_quick_fill_save_station")

        self.starting_area_quick_fill_layout.addWidget(self.starting_area_quick_fill_save_station)


        self.starting_area_layout.addLayout(self.starting_area_quick_fill_layout)

        self.starting_locations_area = QScrollArea(self.starting_area_tab)
        self.starting_locations_area.setObjectName(u"starting_locations_area")
        self.starting_locations_area.setWidgetResizable(True)
        self.starting_locations_contents = QWidget()
        self.starting_locations_contents.setObjectName(u"starting_locations_contents")
        self.starting_locations_contents.setGeometry(QRect(0, 0, 76, 16))
        self.starting_locations_layout = QGridLayout(self.starting_locations_contents)
        self.starting_locations_layout.setSpacing(6)
        self.starting_locations_layout.setContentsMargins(11, 11, 11, 11)
        self.starting_locations_layout.setObjectName(u"starting_locations_layout")
        self.starting_locations_layout.setContentsMargins(4, 4, 4, 4)
        self.starting_locations_area.setWidget(self.starting_locations_contents)

        self.starting_area_layout.addWidget(self.starting_locations_area)

        self.patches_tab_widget.addTab(self.starting_area_tab, "")

        self.patches_tab_layout.addWidget(self.patches_tab_widget)

        self.main_tab_widget.addTab(self.patches_tab, "")

        self.verticalLayout.addWidget(self.main_tab_widget)

        self.button_box = QDialogButtonBox(LogicSettingsWindow)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Save)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(LogicSettingsWindow)

        self.main_tab_widget.setCurrentIndex(0)
        self.logic_tab_widget.setCurrentIndex(0)
        self.patches_tab_widget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(LogicSettingsWindow)
    # setupUi

    def retranslateUi(self, LogicSettingsWindow):
        LogicSettingsWindow.setWindowTitle(QCoreApplication.translate("LogicSettingsWindow", u"Customize Preset", None))
        self.name_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Name:", None))
        self.logic_description_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p align=\"justify\">Randovania has rules in place which guarantees that the game is completable regardless of the modifications made to the game. Here you can also configure which kind of game knowledge or skill it expects you to have, allowing for even more varied games.</p><p align=\"justify\">No matter the level, it is always possible to softlock when you enter a room or area that you're unable to leave. For example, vanilla beam rooms without the necessary beam to escape, Dark World without Light Beam/Anihhilator Beam, Torvus Bog without Super Missile.</p><p align=\"justify\">However, it may be <span style=\" font-style:italic;\">necessary</span> to enter Dark World without a way to escape if that item is located in the Dark World.</p></body></html>", None))
        self.trick_level_minimal_logic_check.setText(QCoreApplication.translate("LogicSettingsWindow", u"Use minimal logic instead", None))
        self.trick_level_minimal_logic_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Randovania will only check that Screw Attack, Dark Visor and Light Suit won't all be behind Ing Caches and Dark Water, removing the biggest reasons for a pure random layout to be impossible.</p><p>There are no guarantees that a seed will be possible in this case.</p><p><br/></p></body></html>", None))
        self.trick_level_help_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>If you want to tweak the knowledge or skill needed expected in a game, you can configure the level used for each of the tweaks listed below by moving the slider to the apropriate level.</p><p>Press the ? button to see which rooms use that trick on the selected level.</p></body></html>", None))
        self.logic_tab_widget.setTabText(self.logic_tab_widget.indexOf(self.trick_level_tab), QCoreApplication.translate("LogicSettingsWindow", u"Logic Level", None))
        self.damage_strictness_group.setTitle(QCoreApplication.translate("LogicSettingsWindow", u"Logic damage strictness", None))
        self.damage_strictness_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Certain locations, such as rooms without safe zones in Dark Aether or bosses, requires a certain number of energy tanks (or suits).</p><p>This setting controls how much energy the logic will expect you to have to reach these locations.</p></body></html>", None))
        self.damage_strictness_combo.setItemText(0, QCoreApplication.translate("LogicSettingsWindow", u"Strict (1\u00d7)", None))
        self.damage_strictness_combo.setItemText(1, QCoreApplication.translate("LogicSettingsWindow", u"Medium (1.5\u00d7)", None))
        self.damage_strictness_combo.setItemText(2, QCoreApplication.translate("LogicSettingsWindow", u"Lenient (2\u00d7)", None))

        self.damage_strictness_combo.setCurrentText(QCoreApplication.translate("LogicSettingsWindow", u"Strict (1\u00d7)", None))
        self.logic_tab_widget.setTabText(self.logic_tab_widget.indexOf(self.logic_damage_tab), QCoreApplication.translate("LogicSettingsWindow", u"Damage", None))
        self.randomization_mode_group.setTitle(QCoreApplication.translate("LogicSettingsWindow", u"Randomization Mode", None))
        self.randomization_mode_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>This setting controls how Randovania will shuffle items.</p><p><span style=\" font-weight:600;\">Full:</span> All items can be placed in any location.</p><p><span style=\" font-weight:600;\">Major/minor split:</span> Major items (i.e., major upgrades, Energy Tanks, Dark Temple Keys, and Energy Transfer Modules) and minor items (i.e, expansions) will be shuffled separately. Major items in excess of the number of major locations will be placed in minor locations, and vice versa.</p></body></html>", None))
        self.randomization_mode_combo.setItemText(0, QCoreApplication.translate("LogicSettingsWindow", u"Full", None))
        self.randomization_mode_combo.setItemText(1, QCoreApplication.translate("LogicSettingsWindow", u"Major/minor split", None))

        self.excluded_locations_group.setTitle(QCoreApplication.translate("LogicSettingsWindow", u"Available Locations", None))
        self.excluded_locations_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Choose which locations are considered for placing items.</p></body></html>", None))
        self.logic_tab_widget.setTabText(self.logic_tab_widget.indexOf(self.location_pool_tab), QCoreApplication.translate("LogicSettingsWindow", u"Location Pool", None))
        self.main_tab_widget.setTabText(self.main_tab_widget.indexOf(self.logic_tab), QCoreApplication.translate("LogicSettingsWindow", u"Randomizer Logic", None))
        self.energy_tank_box.setTitle(QCoreApplication.translate("LogicSettingsWindow", u"Energy tank", None))
        self.dangerous_tank_check.setText(QCoreApplication.translate("LogicSettingsWindow", u"1 HP mode. In this mode, Energy Tanks and Save Stations leave you at 1 HP instead", None))
        self.energy_tank_capacity_description.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Configure how much energy each Energy Tank you collect will provide. Your base energy is always this quantity, minus 1.</p><p>While logic will respect this value, only the original value (100) has been tested.</p></body></html>", None))
        self.energy_tank_capacity_spin_box.setSuffix(QCoreApplication.translate("LogicSettingsWindow", u" energy", None))
        self.energy_tank_capacity_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Energy per tank", None))
        self.safe_zone_box.setTitle(QCoreApplication.translate("LogicSettingsWindow", u"Safe zone", None))
        self.safe_zone_regen_spin.setSuffix(QCoreApplication.translate("LogicSettingsWindow", u" energy/s", None))
        self.safe_zone_regen_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Safe Zone healing rate", None))
        self.safe_zone_logic_heal_check.setText(QCoreApplication.translate("LogicSettingsWindow", u"Logic considers fully healing at every safe zone. This is currently always on.", None))
        self.safe_zone_description.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Configure how Dark Aether safe zones operate and how logic uses them.</p></body></html>", None))
        self.dark_aether_box.setTitle(QCoreApplication.translate("LogicSettingsWindow", u"Dark Aether", None))
        self.varia_suit_spin_box.setSuffix(QCoreApplication.translate("LogicSettingsWindow", u" energy/s", None))
        self.varia_suit_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Varia Suit", None))
        self.dark_suit_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Dark Suit", None))
        self.dark_suit_spin_box.setSuffix(QCoreApplication.translate("LogicSettingsWindow", u" energy/s", None))
        self.dark_aether_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Configure how much damage per second you take in Dark Aether, per suit.<br/>Light Suit is always imune.</p><p>Logic always use the default value for checking, so higher values may cause impossible games.</p></body></html>", None))
        self.patches_tab_widget.setTabText(self.patches_tab_widget.indexOf(self.energy_tab), QCoreApplication.translate("LogicSettingsWindow", u"Energy", None))
        self.elevators_description_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>Controls where each elevator connects to.</p><p><span style=\" font-weight:600;\">Original Connections</span>: all elevators are connected to where they do in the original game.</p><p><span style=\" font-weight:600;\">Two-way, between areas</span>: after taking an elevator, the elevator in the room you are in will bring you back to where you were. An elevator will never connect to another in the same area.</p><p><span style=\" font-weight:600;\">Two-way, unchecked</span>: after taking an elevator, the elevator in the room you are in will bring you back to where you were. <span style=\" font-style:italic;\">Experimental, can cause incompletable matches.</span></p><p><span style=\" font-weight:600;\">One-way, elevator room</span>: all elevators bring you to another elevator room, but going backwards can go somewhere else. <span style=\" font-style:italic;\">Experimental, can cause incompletable matches.</span></p><p><span style=\" font-weight:600;\">One-way, anywhere</span>: elevators are c"
                        "onnected to any room from the game. <span style=\" font-style:italic;\">Experimental, can cause incompletable matches.</span><br/><span style=\" font-style:italic;\">Warning:</span> there's an audio bug whenever you take an elevator. This bug lasts until you restart the game.</p></body></html>", None))
        self.elevators_combo.setItemText(0, QCoreApplication.translate("LogicSettingsWindow", u"Original Connections", None))
        self.elevators_combo.setItemText(1, QCoreApplication.translate("LogicSettingsWindow", u"Random: Two-way, between areas", None))
        self.elevators_combo.setItemText(2, QCoreApplication.translate("LogicSettingsWindow", u"Random: Two-way, unchecked", None))
        self.elevators_combo.setItemText(3, QCoreApplication.translate("LogicSettingsWindow", u"Random: One-way, elevator room", None))
        self.elevators_combo.setItemText(4, QCoreApplication.translate("LogicSettingsWindow", u"Random: One-way, anywhere", None))

        self.patches_tab_widget.setTabText(self.patches_tab_widget.indexOf(self.elevator_tab), QCoreApplication.translate("LogicSettingsWindow", u"Elevators", None))
        self.startingarea_description.setText(QCoreApplication.translate("LogicSettingsWindow", u"<html><head/><body><p>The area where the game starts at can be customized, being selected randomly from a list.</p><p>For ease of use, you can select some pre-defined list of locations. They are:<br/>Ship: Just Temple Grounds - Landing Site, the vanilla location.<br/>Save Stations: All areas with Save Stations.</p><p><span style=\" font-weight:600;\">WARNING</span>: depending on the starting items that are configured, it may be impossible to start at the chosen place. In that case, the generation will fail.</p></body></html>", None))
        self.starting_area_quick_fill_label.setText(QCoreApplication.translate("LogicSettingsWindow", u"Quick Fill with:", None))
        self.starting_area_quick_fill_ship.setText(QCoreApplication.translate("LogicSettingsWindow", u"Ship (Default)", None))
        self.starting_area_quick_fill_save_station.setText(QCoreApplication.translate("LogicSettingsWindow", u"Save Station", None))
        self.patches_tab_widget.setTabText(self.patches_tab_widget.indexOf(self.starting_area_tab), QCoreApplication.translate("LogicSettingsWindow", u"Starting Area", None))
        self.main_tab_widget.setTabText(self.main_tab_widget.indexOf(self.patches_tab), QCoreApplication.translate("LogicSettingsWindow", u"Game Modifications", None))
    # retranslateUi




import pytest

from randovania.gui.dialog.background_process_dialog import BackgroundProcessDialog


def invoke_callable(target):
    target()


@pytest.mark.asyncio
async def test_open_for_background_task(skip_qtbot, mocker):
    mocker.patch("randovania.gui.dialog.background_process_dialog.BackgroundProcessDialog._start_thread_for",
                 side_effect=invoke_callable)

    def work(progress_update):
        progress_update("Hello", 1)
        return 5

    result = await BackgroundProcessDialog.open_for_background_task(work, "Starting")
    assert result == 5



from PySide2.QtCore import Qt, QEvent
from PySide2.QtWidgets import QSpinBox


class CustomSpinBox(QSpinBox):
    def __init__(self, parent):
        super().__init__(parent)
        self.installEventFilter(self)
        self.setFocusPolicy(Qt.StrongFocus)

    def focusInEvent(self, event: QEvent):
        self.setFocusPolicy(Qt.WheelFocus)

    def focusOutEvent(self, event: QEvent):
        self.setFocusPolicy(Qt.StrongFocus)

    def eventFilter(self, obj: QSpinBox, event: QEvent) -> bool:
        if event.type() == QEvent.Wheel and isinstance(obj, QSpinBox):
            if obj.focusPolicy() == Qt.WheelFocus:
                event.accept()
                return False
            else:
                event.ignore()
                return True
        return super().eventFilter(obj, event)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'echoes_user_preferences_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_EchoesUserPreferencesDialog(object):
    def setupUi(self, EchoesUserPreferencesDialog):
        if not EchoesUserPreferencesDialog.objectName():
            EchoesUserPreferencesDialog.setObjectName(u"EchoesUserPreferencesDialog")
        EchoesUserPreferencesDialog.resize(424, 421)
        self.gridLayout = QGridLayout(EchoesUserPreferencesDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.reset_button = QPushButton(EchoesUserPreferencesDialog)
        self.reset_button.setObjectName(u"reset_button")

        self.gridLayout.addWidget(self.reset_button, 2, 2, 1, 1)

        self.accept_button = QPushButton(EchoesUserPreferencesDialog)
        self.accept_button.setObjectName(u"accept_button")

        self.gridLayout.addWidget(self.accept_button, 2, 0, 1, 1)

        self.cancel_button = QPushButton(EchoesUserPreferencesDialog)
        self.cancel_button.setObjectName(u"cancel_button")

        self.gridLayout.addWidget(self.cancel_button, 2, 1, 1, 1)

        self.scrollArea = QScrollArea(EchoesUserPreferencesDialog)
        self.scrollArea.setObjectName(u"scrollArea")
        self.scrollArea.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 390, 695))
        self.verticalLayout = QVBoxLayout(self.scroll_area_contents)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.description_label = QLabel(self.scroll_area_contents)
        self.description_label.setObjectName(u"description_label")
        self.description_label.setWordWrap(True)

        self.verticalLayout.addWidget(self.description_label)

        self.game_changes_box = QGroupBox(self.scroll_area_contents)
        self.game_changes_box.setObjectName(u"game_changes_box")
        self.game_changes_layout = QVBoxLayout(self.game_changes_box)
        self.game_changes_layout.setSpacing(6)
        self.game_changes_layout.setContentsMargins(11, 11, 11, 11)
        self.game_changes_layout.setObjectName(u"game_changes_layout")
        self.faster_credits_check = QCheckBox(self.game_changes_box)
        self.faster_credits_check.setObjectName(u"faster_credits_check")

        self.game_changes_layout.addWidget(self.faster_credits_check)

        self.remove_hud_popup_check = QCheckBox(self.game_changes_box)
        self.remove_hud_popup_check.setObjectName(u"remove_hud_popup_check")

        self.game_changes_layout.addWidget(self.remove_hud_popup_check)

        self.open_map_check = QCheckBox(self.game_changes_box)
        self.open_map_check.setObjectName(u"open_map_check")

        self.game_changes_layout.addWidget(self.open_map_check)

        self.pickup_markers_check = QCheckBox(self.game_changes_box)
        self.pickup_markers_check.setObjectName(u"pickup_markers_check")

        self.game_changes_layout.addWidget(self.pickup_markers_check)


        self.verticalLayout.addWidget(self.game_changes_box)

        self.visor_box = QGroupBox(self.scroll_area_contents)
        self.visor_box.setObjectName(u"visor_box")
        self.visor_layout = QGridLayout(self.visor_box)
        self.visor_layout.setSpacing(6)
        self.visor_layout.setContentsMargins(11, 11, 11, 11)
        self.visor_layout.setObjectName(u"visor_layout")
        self.hud_lag_check = QCheckBox(self.visor_box)
        self.hud_lag_check.setObjectName(u"hud_lag_check")

        self.visor_layout.addWidget(self.hud_lag_check, 3, 0, 1, 2)

        self.hud_alpha_label = QLabel(self.visor_box)
        self.hud_alpha_label.setObjectName(u"hud_alpha_label")

        self.visor_layout.addWidget(self.hud_alpha_label, 0, 0, 1, 1)

        self.helmet_alpha_slider = QSlider(self.visor_box)
        self.helmet_alpha_slider.setObjectName(u"helmet_alpha_slider")
        self.helmet_alpha_slider.setOrientation(Qt.Horizontal)
        self.helmet_alpha_slider.setTickPosition(QSlider.TicksBelow)

        self.visor_layout.addWidget(self.helmet_alpha_slider, 1, 1, 1, 1)

        self.hud_alpha_value_label = QLabel(self.visor_box)
        self.hud_alpha_value_label.setObjectName(u"hud_alpha_value_label")

        self.visor_layout.addWidget(self.hud_alpha_value_label, 0, 2, 1, 1)

        self.helmet_alpha_label = QLabel(self.visor_box)
        self.helmet_alpha_label.setObjectName(u"helmet_alpha_label")

        self.visor_layout.addWidget(self.helmet_alpha_label, 1, 0, 1, 1)

        self.helmet_alpha_value_label = QLabel(self.visor_box)
        self.helmet_alpha_value_label.setObjectName(u"helmet_alpha_value_label")

        self.visor_layout.addWidget(self.helmet_alpha_value_label, 1, 2, 1, 1)

        self.hud_alpha_slider = QSlider(self.visor_box)
        self.hud_alpha_slider.setObjectName(u"hud_alpha_slider")
        self.hud_alpha_slider.setOrientation(Qt.Horizontal)
        self.hud_alpha_slider.setTickPosition(QSlider.TicksBelow)

        self.visor_layout.addWidget(self.hud_alpha_slider, 0, 1, 1, 1)

        self.checkBox = QCheckBox(self.visor_box)
        self.checkBox.setObjectName(u"checkBox")
        self.checkBox.setEnabled(False)

        self.visor_layout.addWidget(self.checkBox, 4, 0, 1, 2)


        self.verticalLayout.addWidget(self.visor_box)

        self.controls_box = QGroupBox(self.scroll_area_contents)
        self.controls_box.setObjectName(u"controls_box")
        self.controls_layout = QGridLayout(self.controls_box)
        self.controls_layout.setSpacing(6)
        self.controls_layout.setContentsMargins(11, 11, 11, 11)
        self.controls_layout.setObjectName(u"controls_layout")
        self.invert_y_axis_check = QCheckBox(self.controls_box)
        self.invert_y_axis_check.setObjectName(u"invert_y_axis_check")

        self.controls_layout.addWidget(self.invert_y_axis_check, 0, 0, 1, 1)

        self.rumble_check = QCheckBox(self.controls_box)
        self.rumble_check.setObjectName(u"rumble_check")

        self.controls_layout.addWidget(self.rumble_check, 1, 0, 1, 1)


        self.verticalLayout.addWidget(self.controls_box)

        self.audio_box = QGroupBox(self.scroll_area_contents)
        self.audio_box.setObjectName(u"audio_box")
        self.audio_layout = QGridLayout(self.audio_box)
        self.audio_layout.setSpacing(6)
        self.audio_layout.setContentsMargins(11, 11, 11, 11)
        self.audio_layout.setObjectName(u"audio_layout")
        self.sound_mode_label = QLabel(self.audio_box)
        self.sound_mode_label.setObjectName(u"sound_mode_label")
        self.sound_mode_label.setMaximumSize(QSize(16777215, 20))

        self.audio_layout.addWidget(self.sound_mode_label, 0, 0, 1, 1)

        self.sfx_volume_label = QLabel(self.audio_box)
        self.sfx_volume_label.setObjectName(u"sfx_volume_label")

        self.audio_layout.addWidget(self.sfx_volume_label, 1, 0, 1, 1)

        self.music_volume_label = QLabel(self.audio_box)
        self.music_volume_label.setObjectName(u"music_volume_label")

        self.audio_layout.addWidget(self.music_volume_label, 2, 0, 1, 1)

        self.sound_mode_combo = QComboBox(self.audio_box)
        self.sound_mode_combo.setObjectName(u"sound_mode_combo")

        self.audio_layout.addWidget(self.sound_mode_combo, 0, 1, 1, 1)

        self.sfx_volume_slider = QSlider(self.audio_box)
        self.sfx_volume_slider.setObjectName(u"sfx_volume_slider")
        self.sfx_volume_slider.setOrientation(Qt.Horizontal)
        self.sfx_volume_slider.setTickPosition(QSlider.TicksBelow)

        self.audio_layout.addWidget(self.sfx_volume_slider, 1, 1, 1, 1)

        self.music_volume_slider = QSlider(self.audio_box)
        self.music_volume_slider.setObjectName(u"music_volume_slider")
        self.music_volume_slider.setOrientation(Qt.Horizontal)
        self.music_volume_slider.setTickPosition(QSlider.TicksBelow)

        self.audio_layout.addWidget(self.music_volume_slider, 2, 1, 1, 1)

        self.sfx_volume_value_label = QLabel(self.audio_box)
        self.sfx_volume_value_label.setObjectName(u"sfx_volume_value_label")

        self.audio_layout.addWidget(self.sfx_volume_value_label, 1, 2, 1, 1)

        self.music_volume_value_label = QLabel(self.audio_box)
        self.music_volume_value_label.setObjectName(u"music_volume_value_label")

        self.audio_layout.addWidget(self.music_volume_value_label, 2, 2, 1, 1)


        self.verticalLayout.addWidget(self.audio_box)

        self.screen_box = QGroupBox(self.scroll_area_contents)
        self.screen_box.setObjectName(u"screen_box")
        self.screen_layout = QGridLayout(self.screen_box)
        self.screen_layout.setSpacing(6)
        self.screen_layout.setContentsMargins(11, 11, 11, 11)
        self.screen_layout.setObjectName(u"screen_layout")
        self.screen_brightness_label = QLabel(self.screen_box)
        self.screen_brightness_label.setObjectName(u"screen_brightness_label")

        self.screen_layout.addWidget(self.screen_brightness_label, 0, 0, 1, 1)

        self.screen_x_offset_label = QLabel(self.screen_box)
        self.screen_x_offset_label.setObjectName(u"screen_x_offset_label")

        self.screen_layout.addWidget(self.screen_x_offset_label, 1, 0, 1, 1)

        self.screen_brightness_slider = QSlider(self.screen_box)
        self.screen_brightness_slider.setObjectName(u"screen_brightness_slider")
        self.screen_brightness_slider.setOrientation(Qt.Horizontal)
        self.screen_brightness_slider.setTickPosition(QSlider.TicksBelow)

        self.screen_layout.addWidget(self.screen_brightness_slider, 0, 1, 1, 1)

        self.screen_y_offset_slider = QSlider(self.screen_box)
        self.screen_y_offset_slider.setObjectName(u"screen_y_offset_slider")
        self.screen_y_offset_slider.setOrientation(Qt.Horizontal)
        self.screen_y_offset_slider.setTickPosition(QSlider.TicksBelow)

        self.screen_layout.addWidget(self.screen_y_offset_slider, 2, 1, 1, 1)

        self.screen_stretch_label = QLabel(self.screen_box)
        self.screen_stretch_label.setObjectName(u"screen_stretch_label")

        self.screen_layout.addWidget(self.screen_stretch_label, 3, 0, 1, 1)

        self.screen_x_offset_slider = QSlider(self.screen_box)
        self.screen_x_offset_slider.setObjectName(u"screen_x_offset_slider")
        self.screen_x_offset_slider.setOrientation(Qt.Horizontal)
        self.screen_x_offset_slider.setTickPosition(QSlider.TicksBelow)

        self.screen_layout.addWidget(self.screen_x_offset_slider, 1, 1, 1, 1)

        self.screen_stretch_slider = QSlider(self.screen_box)
        self.screen_stretch_slider.setObjectName(u"screen_stretch_slider")
        self.screen_stretch_slider.setOrientation(Qt.Horizontal)
        self.screen_stretch_slider.setTickPosition(QSlider.TicksBelow)

        self.screen_layout.addWidget(self.screen_stretch_slider, 3, 1, 1, 1)

        self.screen_y_offset_label = QLabel(self.screen_box)
        self.screen_y_offset_label.setObjectName(u"screen_y_offset_label")

        self.screen_layout.addWidget(self.screen_y_offset_label, 2, 0, 1, 1)

        self.screen_brightness_value_label = QLabel(self.screen_box)
        self.screen_brightness_value_label.setObjectName(u"screen_brightness_value_label")

        self.screen_layout.addWidget(self.screen_brightness_value_label, 0, 2, 1, 1)

        self.screen_x_offset_value_label = QLabel(self.screen_box)
        self.screen_x_offset_value_label.setObjectName(u"screen_x_offset_value_label")

        self.screen_layout.addWidget(self.screen_x_offset_value_label, 1, 2, 1, 1)

        self.screen_y_offset_value_label = QLabel(self.screen_box)
        self.screen_y_offset_value_label.setObjectName(u"screen_y_offset_value_label")

        self.screen_layout.addWidget(self.screen_y_offset_value_label, 2, 2, 1, 1)

        self.screen_stretch_value_label = QLabel(self.screen_box)
        self.screen_stretch_value_label.setObjectName(u"screen_stretch_value_label")

        self.screen_layout.addWidget(self.screen_stretch_value_label, 3, 2, 1, 1)


        self.verticalLayout.addWidget(self.screen_box)

        self.scrollArea.setWidget(self.scroll_area_contents)

        self.gridLayout.addWidget(self.scrollArea, 1, 0, 1, 3)


        self.retranslateUi(EchoesUserPreferencesDialog)

        QMetaObject.connectSlotsByName(EchoesUserPreferencesDialog)
    # setupUi

    def retranslateUi(self, EchoesUserPreferencesDialog):
        EchoesUserPreferencesDialog.setWindowTitle(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Game Patching", None))
        self.reset_button.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Reset to Defaults", None))
        self.accept_button.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Accept", None))
        self.cancel_button.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Cancel", None))
        self.description_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"<html><head/><body><p>What you choose here will be used as the default values for the in-game options.</p></body></html>", None))
        self.game_changes_box.setTitle(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Game Changes", None))
        self.faster_credits_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Faster Credits", None))
        self.remove_hud_popup_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Skip Item Acquisition Popups", None))
        self.open_map_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Open map from start", None))
        self.pickup_markers_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Replace Translator icons on map with item icons", None))
        self.visor_box.setTitle(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Visor", None))
        self.hud_lag_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Hud Lag", None))
        self.hud_alpha_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Visor Opacity", None))
        self.hud_alpha_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
        self.helmet_alpha_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Helmet Opacity", None))
        self.helmet_alpha_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
#if QT_CONFIG(tooltip)
        self.checkBox.setToolTip(QCoreApplication.translate("EchoesUserPreferencesDialog", u"<html><head/><body><p>The in-game Hint System has been removed. The option for it remains, but does nothing.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.checkBox.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Hint System", None))
        self.controls_box.setTitle(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Controls", None))
        self.invert_y_axis_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Invert Y Axis", None))
        self.rumble_check.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Rumble", None))
        self.audio_box.setTitle(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Audio", None))
        self.sound_mode_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Sound Mode", None))
        self.sfx_volume_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Sound Volume", None))
        self.music_volume_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Music Volume", None))
        self.sfx_volume_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
        self.music_volume_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
        self.screen_box.setTitle(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Screen", None))
        self.screen_brightness_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Screen Brightness", None))
        self.screen_x_offset_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Screen X Offset", None))
        self.screen_stretch_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Screen Stretch", None))
        self.screen_y_offset_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"Screen Y Offset", None))
        self.screen_brightness_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
        self.screen_x_offset_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
        self.screen_y_offset_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
        self.screen_stretch_value_label.setText(QCoreApplication.translate("EchoesUserPreferencesDialog", u"TextLabel", None))
    # retranslateUi




import copy

import pytest

from randovania.game_description.data_reader import WorldReader
from randovania.game_description.resources.resource_database import ResourceDatabase, MissingResource


def test_copy_worlds(echoes_game_description):
    game_copy = copy.deepcopy(echoes_game_description)

    assert echoes_game_description.world_list.worlds == game_copy.world_list.worlds
    assert echoes_game_description.world_list.worlds is not game_copy.world_list.worlds


def test_invalid_node_type():
    # Setup
    reader = WorldReader(None, None)

    with pytest.raises(Exception) as e:
        reader.read_node({
            "name": "Broken Node",
            "heal": True,
            "coordinates": None,
            "node_type": "something that doesn't exist"
        })

    assert str(e.value) == "In node Broken Node, got error: Unknown type: something that doesn't exist"


def test_area_with_invalid_connections():
    # Setup
    db = ResourceDatabase([], [], [], [], [], [], {}, 0, 0, 0)
    reader = WorldReader(db, None)

    with pytest.raises(MissingResource) as e:
        reader.read_area({
            "name": "Broken Area",
            "nodes": [
                {"name": "A", "heal": True, "coordinates": None, "node_type": "generic", "connections": {}},
                {"name": "Broken", "heal": True, "coordinates": None, "node_type": "generic", "connections": {
                    "A": {
                        "type": "resource",
                        "data": {
                            "type": 0,
                            "index": 1,
                            "amount": 1,
                            "negate": False
                        }
                    }
                }},
            ]
        })

    assert str(e.value) == ("In area Broken Area, connection from Broken to A got error: "
                            "ResourceType.ITEM Resource with index 1 not found in []")



import pytest

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.major_item import MajorItem
from randovania.layout.major_item_state import MajorItemState


@pytest.fixture(
    params=[
        {"encoded": b'\x04', "json": {}},
        {"encoded": b'\x14', "json": {"num_shuffled_pickups": 1}},
        {"encoded": b'$', "json": {"num_shuffled_pickups": 2}},
        {"encoded": b'4', "json": {"num_shuffled_pickups": 3}},
        {"encoded": b'Me', "json": {"num_shuffled_pickups": 99}},

        {"encoded": b'EP', "category": "energy_tank", "json": {"num_shuffled_pickups": 6,
                                                               "num_included_in_starting_items": 10}},

        {"encoded": b'\x076D', "ammo_index": (10, 20), "json": {"included_ammo": [230, 200]}},
    ],
    name="state_with_data")
def _state_with_data(request):
    item = MajorItem(
        name="Item Name",
        item_category=ItemCategory(request.param.get("category", "visor")),
        broad_category=ItemCategory(request.param.get("broad_category", "visor")),
        model_index=0,
        progression=(),
        ammo_index=request.param.get("ammo_index", ()),
        required=True,
        original_index=None,
        probability_offset=0,
    )
    return item, request.param["encoded"], MajorItemState.from_json(request.param["json"])


def test_decode(state_with_data):
    # Setup
    item, data, expected = state_with_data

    # Run
    decoder = BitPackDecoder(data)
    result = MajorItemState.bit_pack_unpack(decoder, item)

    # Assert
    assert result == expected


def test_encode(state_with_data):
    # Setup
    item, expected, value = state_with_data

    # Run
    result = bitpacking._pack_encode_results([
        (value_argument, value_format)
        for value_argument, value_format in value.bit_pack_encode(item)
    ])

    # Assert
    assert result == expected


def test_blank_as_json():
    assert MajorItemState().as_json == {}


def test_blank_from_json():
    assert MajorItemState.from_json({}) == MajorItemState()



from typing import Optional, Type


class BaseNetworkError(Exception):

    @classmethod
    def code(cls):
        return NotImplementedError()

    @property
    def detail(self):
        return None

    @property
    def as_json(self):
        return {
            "error": {
                "code": self.code(),
                "detail": self.detail,
            }
        }

    @classmethod
    def from_detail(cls, detail) -> "BaseNetworkError":
        return cls()


class NotLoggedIn(BaseNetworkError):
    @classmethod
    def code(cls):
        return 1


class WrongPassword(BaseNetworkError):
    @classmethod
    def code(cls):
        return 2


class NotAuthorizedForAction(BaseNetworkError):
    @classmethod
    def code(cls):
        return 3


class InvalidAction(BaseNetworkError):
    def __init__(self, message: str):
        self.message = message

    @classmethod
    def code(cls):
        return 4

    @property
    def detail(self):
        return self.message

    @classmethod
    def from_detail(cls, detail) -> "InvalidAction":
        return cls(detail)

    def __str__(self):
        return f"Invalid Action: {self.message}"


class InvalidSession(BaseNetworkError):
    @classmethod
    def code(cls):
        return 5


class ServerError(BaseNetworkError):
    @classmethod
    def code(cls):
        return 6


def decode_error(data: dict) -> Optional[BaseNetworkError]:
    if "error" not in data:
        return None

    code = data["error"]["code"]
    detail = data["error"]["detail"]

    for cls in BaseNetworkError.__subclasses__():
        cls: Type[BaseNetworkError] = cls
        if code == cls.code():
            return cls.from_detail(detail)

    raise RuntimeError("Unknown error")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'background_process_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_BackgroundProcessDialog(object):
    def setupUi(self, BackgroundProcessDialog):
        if not BackgroundProcessDialog.objectName():
            BackgroundProcessDialog.setObjectName(u"BackgroundProcessDialog")
        BackgroundProcessDialog.resize(449, 80)
        self.gridLayout = QGridLayout(BackgroundProcessDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.update_button = QPushButton(BackgroundProcessDialog)
        self.update_button.setObjectName(u"update_button")

        self.gridLayout.addWidget(self.update_button, 1, 1, 1, 1)

        self.progress_bar = QProgressBar(BackgroundProcessDialog)
        self.progress_bar.setObjectName(u"progress_bar")
        self.progress_bar.setValue(0)
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setInvertedAppearance(False)

        self.gridLayout.addWidget(self.progress_bar, 1, 0, 1, 1)

        self.progress_label = QLabel(BackgroundProcessDialog)
        self.progress_label.setObjectName(u"progress_label")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progress_label.sizePolicy().hasHeightForWidth())
        self.progress_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(7)
        self.progress_label.setFont(font)
        self.progress_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.progress_label.setWordWrap(True)

        self.gridLayout.addWidget(self.progress_label, 0, 0, 1, 3)


        self.retranslateUi(BackgroundProcessDialog)

        QMetaObject.connectSlotsByName(BackgroundProcessDialog)
    # setupUi

    def retranslateUi(self, BackgroundProcessDialog):
        BackgroundProcessDialog.setWindowTitle(QCoreApplication.translate("BackgroundProcessDialog", u"Import permalink", None))
        self.update_button.setText(QCoreApplication.translate("BackgroundProcessDialog", u"Stop", None))
        self.progress_label.setText(QCoreApplication.translate("BackgroundProcessDialog", u"<empty>", None))
    # retranslateUi




from typing import Optional, Tuple, List

from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.major_item import MajorItem
from randovania.game_description.resources.pickup_entry import ConditionalResources, ResourceConversion, PickupEntry
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceQuantity
from randovania.games.prime import corruption_items
from randovania.games.prime import prime_items
from randovania.games.prime.echoes_items import DARK_TEMPLE_KEY_MODEL, DARK_TEMPLE_KEY_NAMES, DARK_TEMPLE_KEY_ITEMS, \
    SKY_TEMPLE_KEY_MODEL, SKY_TEMPLE_KEY_ITEMS, USELESS_PICKUP_MODEL, USELESS_PICKUP_ITEM
from randovania.layout.major_item_state import MajorItemState
from randovania.resolver.exceptions import InvalidConfiguration


def create_major_item(item: MajorItem,
                      state: MajorItemState,
                      include_percentage: bool,
                      resource_database: ResourceDatabase,
                      ammo: Optional[Ammo],
                      ammo_requires_major_item: bool,

                      ) -> PickupEntry:
    """
    Creates a Pickup for the given MajorItem
    :param include_percentage:
    :param state:
    :param item:
    :param resource_database:
    :param ammo:
    :param ammo_requires_major_item:
    :return:
    """

    def _create_resources(base_resource: Optional[int],
                          temporary_ammo: bool = False) -> Tuple[ResourceQuantity, ...]:
        resources = []

        if base_resource is not None:
            # FIXME: hacky quantity for Hazard Shield
            quantity = 5 if item.name == "Hazard Shield" else 1
            resources.append((resource_database.get_item(base_resource), quantity))

        for ammo_index, ammo_count in zip(ammo.temporaries if temporary_ammo else item.ammo_index,
                                          state.included_ammo):
            resources.append((resource_database.get_item(ammo_index), ammo_count))

        if include_percentage:
            resources.append((resource_database.item_percentage, 1))

        return tuple(resources)

    if item.progression:
        if ammo_requires_major_item and ammo.unlocked_by != item.progression[0] and ammo.unlocked_by is not None:
            if len(item.progression) != 1:
                raise InvalidConfiguration(
                    ("Item {item.name} uses ammo {ammo.name} that is locked behind {ammo.unlocked_by},"
                     "but it also has progression. This is unsupported.").format(
                        ammo=ammo,
                        item=item,
                    )
                )

            name = resource_database.get_item(item.progression[0]).long_name
            conditional_resources = (
                ConditionalResources(
                    name=name,
                    item=None,
                    resources=_create_resources(item.progression[0], True)
                ),
                ConditionalResources(
                    name=name,
                    item=resource_database.get_item(ammo.unlocked_by),
                    resources=_create_resources(item.progression[0])
                )
            )
        else:
            conditional_resources = tuple(
                ConditionalResources(
                    name=resource_database.get_item(item.progression[i]).long_name,
                    item=resource_database.get_item(item.progression[i - 1]) if i > 0 else None,
                    resources=_create_resources(progression)
                )
                for i, progression in enumerate(item.progression)
            )
    else:
        conditional_resources = (
            ConditionalResources(name=item.name, item=None, resources=_create_resources(None)),
        )

    if item.converts_indices:
        assert len(item.converts_indices) == len(item.ammo_index)
        convert_resources = tuple(
            ResourceConversion(
                source=resource_database.get_item(source),
                target=resource_database.get_item(target),
            )
            for source, target in zip(item.converts_indices, item.ammo_index)
        )
    else:
        convert_resources = tuple()

    return PickupEntry(
        name=item.name,
        resources=conditional_resources,
        model_index=item.model_index,
        item_category=item.item_category,
        broad_category=item.broad_category,
        probability_offset=item.probability_offset,
        probability_multiplier=item.probability_multiplier,
        convert_resources=convert_resources,
    )


def create_ammo_expansion(ammo: Ammo,
                          ammo_count: List[int],
                          requires_major_item: bool,
                          resource_database: ResourceDatabase,
                          ) -> PickupEntry:
    """
    Creates a Pickup for an expansion of the given ammo.
    :param ammo:
    :param ammo_count:
    :param requires_major_item:
    :param resource_database:
    :return:
    """
    resources = [(resource_database.get_item(item), count)
                 for item, count in zip(ammo.items, ammo_count)]
    resources.append((resource_database.item_percentage, 1))

    if ammo.unlocked_by is not None and requires_major_item:
        temporary_resources = [(resource_database.get_item(item), count)
                               for item, count in zip(ammo.temporaries, ammo_count)]
        temporary_resources.append((resource_database.item_percentage, 1))

        conditional_resources = (
            ConditionalResources(temporary_resources[0][0].long_name, None, tuple(temporary_resources)),
            ConditionalResources(ammo.name, resource_database.get_item(ammo.unlocked_by), tuple(resources)),
        )
    else:
        conditional_resources = (
            ConditionalResources(None, None, tuple(resources)),
        )

    return PickupEntry(
        name=ammo.name,
        resources=conditional_resources,
        model_index=ammo.models[0],  # TODO: use a random model
        item_category=ItemCategory.EXPANSION,
        broad_category=ammo.broad_category,
    )


def create_dark_temple_key(key_number: int,
                           temple_index: int,
                           resource_database: ResourceDatabase,
                           ) -> PickupEntry:
    """
    Creates a Dark Temple Key
    :param key_number:
    :param temple_index: The index of the temple: Dark Agon, Dark Torvus, Hive Temple
    :param resource_database:
    :return:
    """

    return PickupEntry(
        name=DARK_TEMPLE_KEY_NAMES[temple_index].format(key_number + 1),
        resources=(
            ConditionalResources(None, None, tuple([
                (resource_database.get_item(DARK_TEMPLE_KEY_ITEMS[temple_index][key_number]), 1)
            ])),
        ),
        model_index=DARK_TEMPLE_KEY_MODEL,
        item_category=ItemCategory.TEMPLE_KEY,
        broad_category=ItemCategory.KEY,
        probability_offset=3,
    )


def create_sky_temple_key(key_number: int,
                          resource_database: ResourceDatabase,
                          ) -> PickupEntry:
    """

    :param key_number:
    :param resource_database:
    :return:
    """

    return PickupEntry(
        name="Sky Temple Key {}".format(key_number + 1),
        resources=(
            ConditionalResources(None, None, tuple([
                (resource_database.get_item(SKY_TEMPLE_KEY_ITEMS[key_number]), 1)
            ])),
        ),
        model_index=SKY_TEMPLE_KEY_MODEL,
        item_category=ItemCategory.SKY_TEMPLE_KEY,
        broad_category=ItemCategory.KEY,
        probability_offset=3,
    )


def create_energy_cell(cell_index: int,
                       resource_database: ResourceDatabase,
                       ) -> PickupEntry:
    return PickupEntry(
        name=f"Energy Cell {cell_index + 1}",
        resources=(
            ConditionalResources(None, None, tuple([
                (resource_database.get_item(corruption_items.ENERGY_CELL_ITEMS[cell_index]), 1),
                (resource_database.get_item(corruption_items.ENERGY_CELL_TOTAL_ITEM), 1),
                (resource_database.item_percentage, 1),
            ])),
        ),
        model_index=corruption_items.ENERGY_CELL_MODEL,
        item_category=ItemCategory.TEMPLE_KEY,
        broad_category=ItemCategory.KEY,
        probability_offset=0.25,
    )


def create_artifact(artifact_index: int,
                    resource_database: ResourceDatabase,
                    ) -> PickupEntry:
    return PickupEntry(
        name=prime_items.ARTIFACT_NAMES[artifact_index],
        resources=(
            ConditionalResources(None, None, tuple([
                (resource_database.get_item(prime_items.ARTIFACT_ITEMS[artifact_index]), 1),
                (resource_database.item_percentage, 1)
            ])),
        ),
        model_index=prime_items.ARTIFACT_MODEL[artifact_index],
        item_category=ItemCategory.TEMPLE_KEY,
        broad_category=ItemCategory.KEY,
        probability_offset=0.25,
    )


def create_useless_pickup(resource_database: ResourceDatabase) -> PickupEntry:
    """
    Creates an Energy Transfer Module pickup.
    :param resource_database:
    :return:
    """
    return PickupEntry(
        name="Energy Transfer Module",
        resources=(
            ConditionalResources(None, None, tuple([
                (resource_database.get_item(USELESS_PICKUP_ITEM), 1)
            ])),
        ),
        model_index=USELESS_PICKUP_MODEL,
        item_category=ItemCategory.ETM,
        broad_category=ItemCategory.ETM,
    )


def create_visual_etm() -> PickupEntry:
    """
    Creates an ETM that should only be used as a visual pickup.
    :return:
    """
    return PickupEntry(
        name="Unknown item",
        resources=(
            ConditionalResources(None, None, tuple()),
        ),
        model_index=USELESS_PICKUP_MODEL,
        item_category=ItemCategory.ETM,
        broad_category=ItemCategory.ETM,
    )



from PySide2 import QtWidgets
from asyncqt import asyncSlot

from randovania.game_description.area import Area
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.dock import DockType, DockConnection
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import Node, GenericNode, DockNode, PickupNode, TeleporterNode, EventNode, \
    TranslatorGateNode, LogbookNode, LoreType, NodeLocation, PlayerShipNode
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import find_resource_info_with_long_name
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world import World
from randovania.gui.dialog.connections_editor import ConnectionsEditor
from randovania.gui.generated.node_details_popup_ui import Ui_NodeDetailsPopup
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.connections_visualizer import ConnectionsVisualizer
from randovania.interface_common import enum_lib


def refresh_if_needed(combo: QtWidgets.QComboBox, func):
    if combo.currentIndex() == 0:
        func(0)


class NodeDetailsPopup(QtWidgets.QDialog, Ui_NodeDetailsPopup):
    _unlocked_by_requirement = Requirement.trivial()
    _connections_visualizer = None

    def __init__(self, game: GameDescription, node: Node):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self.game = game
        self.node = node
        self.world = game.world_list.nodes_to_world(node)
        world = self.world

        self._type_to_tab = {
            GenericNode: self.tab_generic,
            DockNode: self.tab_dock,
            PickupNode: self.tab_pickup,
            TeleporterNode: self.tab_teleporter,
            EventNode: self.tab_event,
            TranslatorGateNode: self.tab_translator_gate,
            LogbookNode: self.tab_logbook,
            PlayerShipNode: self.tab_player_ship,
        }
        tab_to_type = {tab: node_type for node_type, tab in self._type_to_tab.items()}

        # Dynamic Stuff
        for i, node_type in enumerate(self._type_to_tab.keys()):
            self.node_type_combo.setItemData(i, node_type)

        for area in world.areas:
            self.dock_connection_area_combo.addItem(area.name, area)
        refresh_if_needed(self.dock_connection_area_combo, self.on_dock_connection_area_combo)

        for i, enum in enumerate(enum_lib.iterate_enum(DockType)):
            self.dock_type_combo.setItemData(i, enum)

        for world in sorted(game.world_list.worlds, key=lambda x: x.name):
            self.teleporter_destination_world_combo.addItem("{0.name} ({0.dark_name})".format(world), userData=world)
        refresh_if_needed(self.teleporter_destination_world_combo, self.on_teleporter_destination_world_combo)

        for event in sorted(game.resource_database.event, key=lambda it: it.long_name):
            self.event_resource_combo.addItem(event.long_name, event)
        if self.event_resource_combo.count() == 0:
            self.event_resource_combo.addItem("No events in database", None)
            self.event_resource_combo.setEnabled(False)

        for i, enum in enumerate(enum_lib.iterate_enum(LoreType)):
            self.lore_type_combo.setItemData(i, enum)
        refresh_if_needed(self.lore_type_combo, self.on_lore_type_combo)

        self.set_unlocked_by(Requirement.trivial())

        # Signals
        self.button_box.accepted.connect(self.try_accept)
        self.button_box.rejected.connect(self.reject)
        self.node_type_combo.currentIndexChanged.connect(self.on_node_type_combo)
        self.dock_connection_area_combo.currentIndexChanged.connect(self.on_dock_connection_area_combo)
        self.dock_connection_node_combo.currentIndexChanged.connect(self.on_dock_connection_node_combo)
        self.dock_type_combo.currentIndexChanged.connect(self.on_dock_type_combo)
        self.teleporter_destination_world_combo.currentIndexChanged.connect(self.on_teleporter_destination_world_combo)
        self.lore_type_combo.currentIndexChanged.connect(self.on_lore_type_combo)
        self.player_ship_unlocked_button.clicked.connect(self.on_player_ship_unlocked_button)

        # Hide the tab bar
        tab_bar: QtWidgets.QTabBar = self.tab_widget.findChild(QtWidgets.QTabBar)
        tab_bar.hide()

        # Values
        self.name_edit.setText(node.name)
        self.heals_check.setChecked(node.heal)
        self.location_group.setChecked(node.location is not None)
        if node.location is not None:
            self.location_x_spin.setValue(node.location.x)
            self.location_y_spin.setValue(node.location.y)
            self.location_z_spin.setValue(node.location.z)

        visible_tab = self._fill_for_type(node)
        self.node_type_combo.setCurrentIndex(self.node_type_combo.findData(tab_to_type[visible_tab]))
        refresh_if_needed(self.node_type_combo, self.on_node_type_combo)

    def _fill_for_type(self, node: Node) -> QtWidgets.QWidget:
        if isinstance(node, GenericNode):
            return self.tab_generic

        elif isinstance(node, DockNode):
            self.fill_for_dock(node)
            return self.tab_dock

        elif isinstance(node, PickupNode):
            self.fill_for_pickup(node)
            return self.tab_pickup

        elif isinstance(node, TeleporterNode):
            self.fill_for_teleporter(node)
            return self.tab_teleporter

        elif isinstance(node, EventNode):
            self.fill_for_event(node)
            return self.tab_event

        elif isinstance(node, TranslatorGateNode):
            self.fill_for_translator_gate(node)
            return self.tab_translator_gate

        elif isinstance(node, LogbookNode):
            self.fill_for_logbook_node(node)
            return self.tab_logbook

        elif isinstance(node, PlayerShipNode):
            self.fill_for_player_ship_node(node)
            return self.tab_player_ship

        else:
            raise ValueError(f"Unknown node type: {node}")

    def fill_for_dock(self, node: DockNode):
        self.dock_index_spin.setValue(node.dock_index)

        # Connection
        other_area = self.game.world_list.area_by_area_location(AreaLocation(self.world.world_asset_id,
                                                                             node.default_connection.area_asset_id))
        self.dock_connection_area_combo.setCurrentIndex(self.dock_connection_area_combo.findData(other_area))
        refresh_if_needed(self.dock_connection_area_combo, self.on_dock_connection_area_combo)
        self.dock_connection_node_combo.setCurrentIndex(
            self.dock_connection_node_combo.findData(node.default_connection.dock_index))
        self.dock_connection_index_raw_spin.setValue(node.default_connection.dock_index)

        # Dock Weakness
        self.dock_type_combo.setCurrentIndex(self.dock_type_combo.findData(node.default_dock_weakness.dock_type))
        refresh_if_needed(self.dock_type_combo, self.on_dock_type_combo)
        self.dock_weakness_combo.setCurrentIndex(self.dock_weakness_combo.findData(node.default_dock_weakness))

    def fill_for_pickup(self, node: PickupNode):
        self.pickup_index_spin.setValue(node.pickup_index.index)
        self.major_location_check.setChecked(node.major_location)

    def fill_for_teleporter(self, node: TeleporterNode):
        world = self.game.world_list.world_by_asset_id(node.default_connection.world_asset_id)
        area = self.game.world_list.area_by_area_location(node.default_connection)

        self.teleporter_instance_id_edit.setText(hex(node.teleporter_instance_id)
                                                 if node.teleporter_instance_id is not None else "")
        self.teleporter_destination_world_combo.setCurrentIndex(self.teleporter_destination_world_combo.findData(world))
        refresh_if_needed(self.teleporter_destination_world_combo, self.on_teleporter_destination_world_combo)
        self.teleporter_destination_area_combo.setCurrentIndex(self.teleporter_destination_area_combo.findData(area))
        self.teleporter_scan_asset_id_edit.setText(hex(node.scan_asset_id) if node.scan_asset_id is not None else "")
        self.teleporter_editable_check.setChecked(node.editable)
        self.teleporter_vanilla_name_edit.setChecked(node.keep_name_when_vanilla)

    def fill_for_event(self, node: EventNode):
        self.event_resource_combo.setCurrentIndex(self.event_resource_combo.findData(node.event))

    def fill_for_translator_gate(self, node: TranslatorGateNode):
        self.translator_gate_spin.setValue(node.gate.index)

    def fill_for_logbook_node(self, node: LogbookNode):
        self.logbook_string_asset_id_edit.setText(hex(node.string_asset_id).upper())
        self.lore_type_combo.setCurrentIndex(self.lore_type_combo.findData(node.lore_type))
        refresh_if_needed(self.lore_type_combo, self.on_lore_type_combo)

        if node.lore_type == LoreType.LUMINOTH_LORE:
            self.logbook_extra_combo.setCurrentIndex(self.logbook_extra_combo.findData(node.required_translator))

        elif node.lore_type == LoreType.LUMINOTH_WARRIOR:
            self.logbook_extra_combo.setCurrentIndex(self.logbook_extra_combo.findData(node.hint_index))

    def fill_for_player_ship_node(self, node: PlayerShipNode):
        self.set_unlocked_by(node.is_unlocked)

    def set_unlocked_by(self, requirement: Requirement):
        if self._connections_visualizer is not None:
            self._connections_visualizer.deleteLater()
            self._connections_visualizer = None

        self._unlocked_by_requirement = requirement
        self._connections_visualizer = ConnectionsVisualizer(
            self.player_ship_unlocked_group,
            self.player_ship_unlocked_layout,
            self.game.resource_database,
            requirement,
            False
        )

    # Signals
    def on_node_type_combo(self, _):
        self.tab_widget.setCurrentWidget(self._type_to_tab[self.node_type_combo.currentData()])

    def on_dock_connection_area_combo(self, _):
        area: Area = self.dock_connection_area_combo.currentData()

        self.dock_connection_node_combo.clear()
        for node in area.nodes:
            if isinstance(node, DockNode):
                self.dock_connection_node_combo.addItem(f"{node.name} - Dock {node.dock_index}", node.dock_index)
        self.dock_connection_node_combo.addItem("Other", None)

    def on_dock_connection_node_combo(self, _):
        self.dock_connection_index_raw_spin.setEnabled(self.dock_connection_node_combo.currentData() is None)

    def on_dock_type_combo(self, _):
        self.dock_weakness_combo.clear()

        for weakness in self.game.dock_weakness_database.get_by_type(self.dock_type_combo.currentData()):
            self.dock_weakness_combo.addItem(weakness.name, weakness)

    def on_teleporter_destination_world_combo(self, _):
        world: World = self.teleporter_destination_world_combo.currentData()

        self.teleporter_destination_area_combo.clear()
        for area in sorted(world.areas, key=lambda x: x.name):
            self.teleporter_destination_area_combo.addItem(area.name, userData=area)

    def on_lore_type_combo(self, _):
        lore_type: LoreType = self.lore_type_combo.currentData()

        self.logbook_extra_combo.clear()

        if lore_type == LoreType.LUMINOTH_LORE:
            self.logbook_extra_label.setText("Translator needed:")
            for item in self.game.resource_database.item:
                self.logbook_extra_combo.addItem(item.long_name, item)

        elif lore_type == LoreType.LUMINOTH_WARRIOR:
            self.logbook_extra_label.setText("Pickup index hinted:")
            for node in self.game.world_list.all_nodes:
                if isinstance(node, PickupNode):
                    self.logbook_extra_combo.addItem("{} - {}".format(
                        self.game.world_list.node_name(node, True, True),
                        node.pickup_index.index,
                    ), node.pickup_index.index)

        else:
            self.logbook_extra_label.setText("Nothing:")
            self.logbook_extra_combo.addItem("Nothing", None)

    @asyncSlot()
    async def on_player_ship_unlocked_button(self):
        self._edit_popup = ConnectionsEditor(self, self.game.resource_database, self._unlocked_by_requirement)
        self._edit_popup.setModal(True)
        try:
            result = await async_dialog.execute_dialog(self._edit_popup)
            if result == QtWidgets.QDialog.Accepted:
                self.set_unlocked_by(self._edit_popup.final_requirement)
        finally:
            self._edit_popup = None

    # Final
    def create_new_node(self) -> Node:
        node_type = self.node_type_combo.currentData()
        name = self.name_edit.text()
        heal = self.heals_check.isChecked()
        location = None
        if self.location_group.isChecked():
            location = NodeLocation(self.location_x_spin.value(),
                                    self.location_y_spin.value(),
                                    self.location_z_spin.value())
        index = self.node.index

        if node_type == GenericNode:
            return GenericNode(name, heal, location, index)

        elif node_type == DockNode:
            return DockNode(
                name, heal, location, index,
                self.dock_index_spin.value(),
                DockConnection(self.dock_connection_area_combo.currentData().area_asset_id,
                               self.dock_connection_node_combo.currentData()),
                self.dock_weakness_combo.currentData(),
            )

        elif node_type == PickupNode:
            return PickupNode(
                name, heal, location, index,
                PickupIndex(self.pickup_index_spin.value()),
                self.major_location_check.isChecked(),
            )

        elif node_type == TeleporterNode:
            instance_id = self.teleporter_instance_id_edit.text()
            scan_asset_id = self.teleporter_scan_asset_id_edit.text()

            return TeleporterNode(
                name, heal, location, index,
                int(instance_id, 0) if instance_id != "" else None,
                AreaLocation(self.teleporter_destination_world_combo.currentData().world_asset_id,
                             self.teleporter_destination_area_combo.currentData().area_asset_id),
                int(scan_asset_id, 0) if scan_asset_id != "" else None,
                self.teleporter_vanilla_name_edit.isChecked(),
                self.teleporter_editable_check.isChecked(),
            )

        elif node_type == EventNode:
            event = self.event_resource_combo.currentData()
            if event is None:
                raise ValueError("There are no events in the database, unable to create EventNode.")
            return EventNode(
                name, heal, location, index,
                event,
            )

        elif node_type == TranslatorGateNode:
            return TranslatorGateNode(
                name, heal, location, index,
                TranslatorGate(self.translator_gate_spin.value()),
                self._get_scan_visor()
            )

        elif node_type == LogbookNode:
            lore_type: LoreType = self.lore_type_combo.currentData()
            if lore_type == LoreType.LUMINOTH_LORE:
                required_translator = self.logbook_extra_combo.currentData()
                if required_translator is None:
                    raise ValueError("Missing required translator.")
            else:
                required_translator = None

            if lore_type == LoreType.LUMINOTH_WARRIOR:
                hint_index = self.logbook_extra_combo.currentData()
            else:
                hint_index = None

            return LogbookNode(
                name, heal, location, index,
                int(self.logbook_string_asset_id_edit.text(), 0),
                self._get_scan_visor(),
                lore_type,
                required_translator,
                hint_index
            )

        elif node_type == PlayerShipNode:
            return PlayerShipNode(
                name, heal, location, index,
                self._unlocked_by_requirement,
                self._get_command_visor()
            )

        else:
            raise RuntimeError(f"Unknown node type: {node_type}")

    def _get_scan_visor(self):
        return find_resource_info_with_long_name(
            self.game.resource_database.item,
            "Scan Visor"
        )

    def _get_command_visor(self):
        return find_resource_info_with_long_name(
            self.game.resource_database.item,
            "Command Visor"
        )

    def try_accept(self):
        try:
            self.create_new_node()
            self.accept()
        except Exception as e:
            QtWidgets.QMessageBox.warning(self, "Invalid configuration",
                                          f"Unable to save node: {e}")



from typing import Tuple
from unittest.mock import MagicMock

import pytest

from randovania.game_description import data_reader
from randovania.game_description.requirements import ResourceRequirement, RequirementList, RequirementSet, \
    RequirementAnd, RequirementOr, Requirement, MAX_DAMAGE, RequirementTemplate
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo


@pytest.fixture(name="database")
def _database() -> ResourceDatabase:
    return ResourceDatabase(
        item=[
            SimpleResourceInfo(0, "A", "A", ""),
            SimpleResourceInfo(1, "B", "B", ""),
            SimpleResourceInfo(2, "C", "C", ""),
        ],
        event=[],
        trick=[],
        damage=[],
        version=[],
        misc=[
            SimpleResourceInfo(0, "Trivial", "Trivial", ""),
            SimpleResourceInfo(1, "Impossible", "Impossible", ""),
        ],
        requirement_template={},
        energy_tank_item_index=0,
        item_percentage_index=0,
        multiworld_magic_item_index=0
    )


def _make_req(name: str):
    req = SimpleResourceInfo(hash(name), name, name, "")
    id_req = ResourceRequirement(req, 1, False)
    return req, id_req


def _req(name: str):
    return _make_req(name)[1]


def make_req_a():
    return _make_req("A")


def make_req_b():
    return _make_req("B")


def make_req_c():
    return _make_req("C")


def make_single_set(id_req: Tuple[SimpleResourceInfo, ResourceRequirement]) -> RequirementSet:
    return RequirementSet([RequirementList([id_req[1]])])


def test_empty_requirement_set_satisfied():
    assert not RequirementSet([]).satisfied({}, 99)


def test_empty_requirement_list_satisfied():
    assert RequirementList([]).satisfied({}, 99)


def test_simplify_requirement_set_static():
    res_a, id_req_a = make_req_a()
    res_b, id_req_b = make_req_b()

    the_set = RequirementOr([
        RequirementAnd([id_req_a]),
        RequirementAnd([id_req_b]),
    ])

    simple_1 = the_set.patch_requirements({res_a: 0, res_b: 0}, 1)
    simple_2 = the_set.patch_requirements({res_a: 0, res_b: 1}, 1)
    simple_3 = the_set.patch_requirements({res_a: 1, res_b: 1}, 1)

    assert simple_1.as_set.alternatives == frozenset()
    assert simple_2.as_set.alternatives == frozenset([RequirementList([])])
    assert simple_3.as_set.alternatives == frozenset([RequirementList([])])


def test_prevent_redundant():
    res_a, id_req_a = make_req_a()
    res_b, id_req_b = make_req_b()

    the_set = RequirementSet([
        RequirementList([id_req_a]),
        RequirementList([id_req_a, id_req_b]),
    ])

    assert the_set.alternatives == frozenset([RequirementList([id_req_a])])


def test_trivial_merge():
    trivial = RequirementSet.trivial()
    impossible = RequirementSet.impossible()
    res_a, id_req_a = make_req_a()

    the_set = RequirementSet([
        RequirementList([id_req_a]),
    ])

    assert trivial.union(trivial) == trivial
    assert trivial.union(the_set) == the_set
    assert the_set.union(trivial) == the_set
    assert trivial.union(impossible) == impossible
    assert impossible.union(the_set) == impossible
    assert the_set.union(impossible) == impossible
    assert the_set.union(the_set) == the_set


@pytest.mark.parametrize(["a", "b", "expected"], [
    (RequirementSet.impossible(), make_single_set(make_req_a()), make_single_set(make_req_a())),
    (RequirementSet.impossible(), RequirementSet.trivial(), RequirementSet.trivial()),
    (RequirementSet.trivial(), make_single_set(make_req_a()), RequirementSet.trivial()),
    (make_single_set(make_req_a()), make_single_set(make_req_b()),
     RequirementSet([RequirementList([make_req_a()[1]]), RequirementList([make_req_b()[1]])])),
])
def test_expand_alternatives(a: RequirementSet, b: RequirementSet, expected: RequirementSet):
    assert a.expand_alternatives(b) == expected


@pytest.mark.parametrize(["input_data", "output_data"], [
    ([], []),
    ([(0, False)], []),
    ([(0, True)], [0]),
    ([(0, True), (0, False)], [0]),
    ([(0, True), (1, False)], [0]),
    ([(0, True), (1, True)], [0, 1]),
])
def test_list_dangerous_resources(input_data, output_data):
    # setup
    req_list = RequirementList(
        (ResourceRequirement(SimpleResourceInfo(item[0], str(item[0]), str(item[0]), ""), 1, item[1])
         for item in input_data))
    expected_result = {
        SimpleResourceInfo(item, str(item), str(item), "")
        for item in output_data
    }

    # run
    result = set(req_list.dangerous_resources)

    # assert
    assert result == expected_result


def test_set_dangerous_resources():
    # setup
    list_a = MagicMock()
    list_b = MagicMock()
    list_a.dangerous_resources = [1, 2, 3]
    list_b.dangerous_resources = ["a", "b", "c"]

    req_set = RequirementSet([])
    req_set.alternatives = frozenset([list_a, list_b])

    # Run
    result = set(req_set.dangerous_resources)

    # Assert
    assert result == {1, 2, 3, "a", "b", "c"}


def test_requirement_as_set_1():
    req = RequirementAnd([
        _req("A"),
        RequirementOr([_req("B"), _req("C")]),
        RequirementOr([_req("D"), _req("E")]),
    ])
    assert req.as_set == RequirementSet([
        RequirementList([_req("A"), _req("B"), _req("D")]),
        RequirementList([_req("A"), _req("B"), _req("E")]),
        RequirementList([_req("A"), _req("C"), _req("D")]),
        RequirementList([_req("A"), _req("C"), _req("E")]),
    ])


def test_requirement_as_set_2():
    req = RequirementAnd([
        Requirement.trivial(),
        _req("A"),
    ])
    assert req.as_set == RequirementSet([
        RequirementList([_req("A")]),
    ])


def test_requirement_as_set_3():
    req = RequirementOr([
        Requirement.impossible(),
        _req("A"),
    ])
    assert req.as_set == RequirementSet([
        RequirementList([_req("A")]),
    ])


def test_requirement_as_set_4():
    req = RequirementOr([
        Requirement.impossible(),
        _req("A"),
        Requirement.trivial(),
    ])
    assert req.as_set == RequirementSet([
        RequirementList([]),
    ])


def test_requirement_as_set_5():
    req = RequirementAnd([
        _req("A"),
        _req("B"),
        _req("C"),
    ])
    assert req.as_set == RequirementSet([
        RequirementList([_req("A"), _req("B"), _req("C")]),
    ])


def test_requirement_and_str():
    req = RequirementAnd([
        _req("B"),
        _req("A"),
        _req("C"),
    ])
    assert str(req) == "(A ≥ 1 and B ≥ 1 and C ≥ 1)"


def test_requirement_or_str():
    req = RequirementOr([
        _req("C"),
        _req("A"),
        _req("B"),
    ])
    assert str(req) == "(A ≥ 1 or B ≥ 1 or C ≥ 1)"


def test_impossible_requirement_as_set():
    assert Requirement.impossible().as_set == RequirementSet.impossible()


def test_impossible_requirement_satisfied():
    assert not Requirement.impossible().satisfied({}, 99)


def test_impossible_requirement_damage():
    assert Requirement.impossible().damage({}) == MAX_DAMAGE


def test_impossible_requirement_str():
    assert str(Requirement.impossible()) == "Impossible"


def test_trivial_requirement_as_set():
    assert Requirement.trivial().as_set == RequirementSet.trivial()


def test_trivial_requirement_satisfied():
    assert Requirement.trivial().satisfied({}, 99)


def test_trivial_requirement_damage():
    assert Requirement.trivial().damage({}) == 0


def test_trivial_requirement_str():
    assert str(Requirement.trivial()) == "Trivial"


@pytest.mark.parametrize(["original", "expected"], [
    (
            RequirementOr([
                RequirementAnd([
                    _req("A"),
                ]),
            ]),
            _req("A"),
    ),
    (
            RequirementAnd([
                RequirementOr([
                    _req("A"),
                ]),
            ]),
            _req("A"),
    ),
    (
            RequirementAnd([
                RequirementOr([
                    _req("B"),
                    Requirement.trivial()
                ]),
                RequirementAnd([
                    Requirement.trivial(),
                    _req("A"),
                ]),
            ]),
            _req("A"),
    ),
    (
            RequirementOr([
                _req("A"),
                RequirementAnd([_req("B"), _req("C")]),
                RequirementAnd([_req("B"), _req("D")]),
            ]),
            RequirementOr([
                _req("A"),
                RequirementAnd([
                    _req("B"),
                    RequirementOr([
                        _req("C"),
                        _req("D"),
                    ]),
                ]),
            ]),
    ),
    (
            RequirementOr([
                RequirementAnd([_req("B"), _req("C")]),
                RequirementAnd([_req("B"), _req("D")]),
            ]),
            RequirementAnd([
                _req("B"),
                RequirementOr([
                    _req("C"),
                    _req("D"),
                ]),
            ]),
    ),
    (
            RequirementOr([
                _req("A"),
                _req("A"),
            ]),
            _req("A"),
    ),
    (
            RequirementAnd([
                _req("A"),
                _req("A"),
            ]),
            _req("A"),
    ),
    (
            RequirementOr([
                RequirementAnd([
                    _req("A"),
                    RequirementOr([
                        _req("A"),
                        RequirementOr([_req("A")])
                    ])
                ]),
                RequirementAnd([
                    _req("A"),
                    RequirementOr([
                        _req("A"),
                        RequirementOr([]),
                    ]),
                ]),
            ]),
            _req("A"),
    ),
    (
            RequirementOr([
                RequirementAnd([
                    _req("A"),
                    RequirementOr([
                        _req("A"),
                        RequirementOr([_req("A")])
                    ])
                ]),
                RequirementAnd([
                    _req("A"),
                    RequirementOr([
                        _req("A"),
                        RequirementOr([_req("A")]),
                    ])
                ])
            ]),
            _req("A"),
    )
])
def test_simplified_requirement(original, expected):
    simplified = original.simplify()
    assert simplified == expected
    assert simplified.as_set == expected.as_set


def test_requirement_template(database):
    # Setup
    database.requirement_template["Use A"] = _make_req("A")[1]
    use_a = RequirementTemplate(database, "Use A")

    # Run
    as_set = use_a.as_set

    # Assert
    assert as_set == make_single_set(_make_req("A"))
    assert hash(use_a)


def test_requirement_template_nested(database):
    # Setup
    use_a = RequirementTemplate(database, "Use A")
    use_b = RequirementTemplate(database, "Use B")

    database.requirement_template["Use A"] = _req("A")
    database.requirement_template["Use B"] = RequirementOr([use_a, _req("B")])

    # Run
    as_set = use_b.as_set

    # Assert
    assert as_set == RequirementSet([
        RequirementList([_req("A")]),
        RequirementList([_req("B")]),
    ])
    assert hash(use_a) != hash(use_b)


def _json_req(amount: int, index: int = 1, resource_type: int = 3):
    return {"type": "resource", "data": {"type": resource_type, "index": index, "amount": amount, "negate": False}}


@pytest.mark.parametrize(["damage", "items", "requirement"], [
    (50, [], {"type": "and", "data": [_json_req(50)]}),
    (MAX_DAMAGE, [], {"type": "and", "data": [_json_req(1, resource_type=0)]}),
    (80, [], {"type": "and", "data": [_json_req(50), _json_req(30)]}),
    (30, [], {"type": "or", "data": [_json_req(50), _json_req(30)]}),
    (50, [], {"type": "or", "data": [_json_req(50), _json_req(1, resource_type=0)]}),
    (0, [1], {"type": "or", "data": [_json_req(50), _json_req(1, resource_type=0)]}),
    (100, [], {"type": "or", "data": [
        _json_req(100),
        {"type": "and", "data": [_json_req(50), _json_req(1, resource_type=0)]},
    ]}),
    (50, [1], {"type": "or", "data": [
        _json_req(100),
        {"type": "and", "data": [_json_req(50), _json_req(1, resource_type=0)]},
    ]}),
    (150, [], {"type": "and", "data": [
        _json_req(100),
        {"type": "or", "data": [_json_req(50), _json_req(1, resource_type=0)]},
    ]}),
    (100, [1], {"type": "and", "data": [
        _json_req(100),
        {"type": "or", "data": [_json_req(50), _json_req(1, resource_type=0)]},
    ]}),
    (200, [], {"type": "and", "data": [_json_req(100), _json_req(100, 2)]}),
    (121, [13], {"type": "and", "data": [_json_req(100), _json_req(100, 2)]}),
    (100, [14], {"type": "and", "data": [_json_req(100), _json_req(100, 2)]}),
])
def test_requirement_damage(damage, items, requirement, echoes_resource_database):
    req = data_reader.read_requirement(requirement, echoes_resource_database)

    resources = {
        echoes_resource_database.get_item(item): 1
        for item in items
    }

    assert req.damage(resources) == damage



import logging
import platform
from abc import abstractmethod


# Reference for this module:
# https://github.com/h3llrais3r/Deluge-PreventSuspendPlus/blob/master/preventsuspendplus/core.py
#
# Copyright (C) 2015 h3llrais3r <pooh_beer_1@hotmail.com>
# Copyright (C) 2009 Ian Martin <ianmartin@cantab.net>


class InhibitorSource:
    @abstractmethod
    def inhibit(self):
        """
        Inhibits sleep until `uninhibit` is called.
        """

    @abstractmethod
    def uninhibit(self):
        """
        Cancels a previous call to `inhibit`.
        """
        pass


class WindowsInhibitorSource(InhibitorSource):
    """https://msdn.microsoft.com/en-us/library/windows/desktop/aa373208(v=vs.85).aspx"""
    ES_CONTINUOUS = 0x80000000
    ES_SYSTEM_REQUIRED = 0x00000001

    def __init__(self):
        import ctypes
        self.ctypes = ctypes

    def inhibit(self):
        self.ctypes.windll.kernel32.SetThreadExecutionState(
            WindowsInhibitorSource.ES_CONTINUOUS | WindowsInhibitorSource.ES_SYSTEM_REQUIRED)

    def uninhibit(self):
        self.ctypes.windll.kernel32.SetThreadExecutionState(WindowsInhibitorSource.ES_CONTINUOUS)


class DummyInhibitorSource(InhibitorSource):
    def inhibit(self):
        pass

    def uninhibit(self):
        pass


class Inhibitor:
    def __init__(self, source: InhibitorSource):
        self.source = source
        self.inhibited = False

    def inhibit(self):
        if not self.inhibited:
            self.source.inhibit()
            self.inhibited = True

    def uninhibit(self):
        if self.inhibited:
            self.source.uninhibit()
            self.inhibited = False

    def __enter__(self):
        self.inhibit()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.uninhibit()


def _get_inhibitor_source() -> InhibitorSource:
    if platform.system() == "Windows":
        try:
            return WindowsInhibitorSource()
        except Exception as e:
            logging.getLogger().warning("Unable to create WindowsInhibitor: {}".format(e))

    return DummyInhibitorSource()


def get_inhibitor() -> Inhibitor:
    return Inhibitor(_get_inhibitor_source())



import asyncio
import copy
import os
import subprocess
from pathlib import Path

_FOLDER = Path(__file__).parent


async def main():
    if os.environ["GITHUB_REF"].startswith("refs/tags/"):
        server_environment = "production"
    else:
        server_environment = "staging"

    remote_host = "randovania.metroidprime.run"
    new_env = copy.copy(os.environ)
    new_env["DOCKER_HOST"] = f"ssh://root@{remote_host}"
    new_env["DOMAIN"] = remote_host
    if server_environment == "production":
        new_env["PATH_PREFIX"] = "randovania"
        new_env["DATA_PATH"] = "/var/randovania/production/data"
        new_env["VERSION"] = os.environ["GITHUB_REF"].split("/")[-1]

    elif server_environment == "staging":
        new_env["PATH_PREFIX"] = "randovania-staging"
        new_env["DATA_PATH"] = "/var/randovania/staging/data"
        new_env["VERSION"] = "sha-{}".format(os.environ["GITHUB_SHA"][:8])
    else:
        raise ValueError(f"Unknown server_environment: {server_environment}")

    subprocess.run(
        [
            "docker-compose",
            "-p", f"randovania-{server_environment}",
            "up",
            "-d",
        ],
        check=True,
        env=new_env,
        cwd=_FOLDER.joinpath("server-docker")
    )


if __name__ == '__main__':
    asyncio.run(main())



import dataclasses
from enum import Enum
from typing import List

from randovania.bitpacking.bitpacking import BitPackDataClass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data
from randovania.layout.ammo_configuration import AmmoConfiguration
from randovania.layout.available_locations import AvailableLocationsConfiguration
from randovania.layout.damage_strictness import LayoutDamageStrictness
from randovania.layout.major_items_configuration import MajorItemsConfiguration
from randovania.layout.pickup_model import PickupModelStyle, PickupModelDataSource
from randovania.layout.starting_location import StartingLocation
from randovania.layout.trick_level import TrickLevelConfiguration


@dataclasses.dataclass(frozen=True)
class BaseConfiguration(BitPackDataClass, JsonDataclass):
    trick_level: TrickLevelConfiguration
    starting_location: StartingLocation
    available_locations: AvailableLocationsConfiguration
    major_items_configuration: MajorItemsConfiguration
    ammo_configuration: AmmoConfiguration
    damage_strictness: LayoutDamageStrictness
    pickup_model_style: PickupModelStyle
    pickup_model_data_source: PickupModelDataSource

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        raise NotImplementedError()

    @property
    def game(self):
        return self.game_enum()

    @property
    def game_data(self) -> dict:
        return default_data.read_json_then_binary(self.game)[1]

    @classmethod
    def from_json(cls, json_dict: dict) -> "BaseConfiguration":
        game = cls.game_enum()
        item_database = default_database.item_database_for_game(game)

        kwargs = {}
        for field in dataclasses.fields(cls):
            try:
                arg = json_dict[field.name]
            except KeyError as e:
                raise KeyError(f"Missing field {e}") from e

            try:
                if issubclass(field.type, Enum):
                    arg = field.type(arg)
                elif hasattr(field.type, "from_json"):
                    extra_args = []
                    if field.name in ("trick_level", "starting_location"):
                        extra_args.append(game)
                    if field.name in ("major_items_configuration", "ammo_configuration"):
                        extra_args.append(item_database)
                    arg = field.type.from_json(arg, *extra_args)
            except ValueError as e:
                raise ValueError(f"Error in field {field.name}: {e}") from e

            kwargs[field.name] = arg

        return cls(**kwargs)

    def dangerous_settings(self) -> List[str]:
        result = []
        for field in dataclasses.fields(self):
            f = getattr(self, field.name)
            if hasattr(f, "dangerous_settings"):
                result.extend(f.dangerous_settings())

        return result



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'login_prompt_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_LoginPromptDialog(object):
    def setupUi(self, LoginPromptDialog):
        if not LoginPromptDialog.objectName():
            LoginPromptDialog.setObjectName(u"LoginPromptDialog")
        LoginPromptDialog.resize(331, 233)
        self.layout = QVBoxLayout(LoginPromptDialog)
        self.layout.setSpacing(6)
        self.layout.setContentsMargins(11, 11, 11, 11)
        self.layout.setObjectName(u"layout")
        self.title_label = QLabel(LoginPromptDialog)
        self.title_label.setObjectName(u"title_label")
        self.title_label.setWordWrap(True)

        self.layout.addWidget(self.title_label)

        self.discord_button = QPushButton(LoginPromptDialog)
        self.discord_button.setObjectName(u"discord_button")

        self.layout.addWidget(self.discord_button)

        self.guest_button = QPushButton(LoginPromptDialog)
        self.guest_button.setObjectName(u"guest_button")

        self.layout.addWidget(self.guest_button)

        self.connection_status_label = QLabel(LoginPromptDialog)
        self.connection_status_label.setObjectName(u"connection_status_label")
        self.connection_status_label.setAlignment(Qt.AlignCenter)

        self.layout.addWidget(self.connection_status_label)

        self.spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.spacer)

        self.privacy_policy_label = QLabel(LoginPromptDialog)
        self.privacy_policy_label.setObjectName(u"privacy_policy_label")
        self.privacy_policy_label.setAlignment(Qt.AlignBottom|Qt.AlignHCenter)
        self.privacy_policy_label.setWordWrap(True)
        self.privacy_policy_label.setOpenExternalLinks(True)

        self.layout.addWidget(self.privacy_policy_label)

        self.button_box = QDialogButtonBox(LoginPromptDialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Ok|QDialogButtonBox.Reset)

        self.layout.addWidget(self.button_box)


        self.retranslateUi(LoginPromptDialog)

        QMetaObject.connectSlotsByName(LoginPromptDialog)
    # setupUi

    def retranslateUi(self, LoginPromptDialog):
        LoginPromptDialog.setWindowTitle(QCoreApplication.translate("LoginPromptDialog", u"Login Window", None))
        self.title_label.setText(QCoreApplication.translate("LoginPromptDialog", u"<html><head/><body><p>Usage of Randovania's online services depends on an user account.</p><p>Currently, only login via Discord is supported.</p></body></html>", None))
        self.discord_button.setText(QCoreApplication.translate("LoginPromptDialog", u"Login with Discord", None))
        self.guest_button.setText(QCoreApplication.translate("LoginPromptDialog", u"Login as Guest", None))
        self.connection_status_label.setText("")
        self.privacy_policy_label.setText(QCoreApplication.translate("LoginPromptDialog", u"<html><head/><body><p>By creating an account, you are agreeing to Randovania's <a href=\"https://github.com/randovania/randovania/blob/master/PRIVACY_POLICY.md\"><span style=\" text-decoration: underline; color:#0000ff;\">Privacy Policy</span></a>.</p></body></html>", None))
    # retranslateUi




import asyncio
from typing import Optional

from PySide2 import QtWidgets

from randovania.gui.lib import common_qt_lib


async def execute_dialog(dialog: QtWidgets.QDialog) -> QtWidgets.QDialog.DialogCode:
    """
    An async version of dialog.exec_, internally using QDialog.show
    :param dialog:
    :return:
    """
    future = asyncio.get_event_loop().create_future()

    dialog.finished.connect(future.set_result)
    try:
        dialog.show()
        return await future
    finally:
        dialog.finished.disconnect(future.set_result)


async def message_box(parent: Optional[QtWidgets.QWidget],
                      icon: QtWidgets.QMessageBox.Icon,
                      title: str, text: str,
                      buttons: QtWidgets.QMessageBox.StandardButtons = QtWidgets.QMessageBox.Ok,
                      default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
                      ) -> QtWidgets.QMessageBox.StandardButton:
    box = QtWidgets.QMessageBox(icon, title, text, buttons, parent)
    box.setDefaultButton(default_button)
    common_qt_lib.set_default_window_icon(box)
    return await execute_dialog(box)


async def warning(parent: Optional[QtWidgets.QWidget],
                  title: str, text: str,
                  buttons: QtWidgets.QMessageBox.StandardButtons = QtWidgets.QMessageBox.Ok,
                  default_button: QtWidgets.QMessageBox.StandardButton = QtWidgets.QMessageBox.NoButton,
                  ) -> QtWidgets.QMessageBox.StandardButton:
    return await message_box(parent, QtWidgets.QMessageBox.Warning, title, text, buttons, default_button)



from PySide2 import QtWidgets, QtGui
from PySide2.QtCore import Signal


class CloseEventWidget(QtWidgets.QMainWindow):
    CloseEvent = Signal()

    def closeEvent(self, event: QtGui.QCloseEvent):
        self.CloseEvent.emit()
        return super().closeEvent(event)



from PySide2 import QtWidgets

from randovania.interface_common.options import Options, DecodeFailedException


def load_options_from_disk(options: Options) -> bool:
    parent: QtWidgets.QWidget = None
    try:
        options.load_from_disk()
        return True

    except DecodeFailedException as decode_failed:
        user_response = QtWidgets.QMessageBox.critical(
            parent,
            "Error loading previous settings",
            ("The following error occurred while restoring your settings:\n"
             "{}\n\n"
             "Do you want to reset this part of your settings?").format(decode_failed),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
            QtWidgets.QMessageBox.No
        )
        if user_response == QtWidgets.QMessageBox.Yes:
            options.load_from_disk(True)
            return True
        else:
            return False



from random import Random
from unittest.mock import MagicMock, patch

import pytest

from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.hint import Hint, HintType, PrecisionPair, HintLocationPrecision, HintItemPrecision, \
    RelativeDataArea, RelativeDataItem
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.node import LogbookNode
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.generator.filler import runner
from randovania.generator.generator import create_player_pool


@patch("randovania.generator.filler.runner.retcon_playthrough_filler", autospec=True)
def test_run_filler(mock_retcon_playthrough_filler: MagicMock,
                    echoes_game_description,
                    default_layout_configuration,
                    ):
    # Setup
    rng = Random(5000)
    status_update = MagicMock()

    logbook_nodes = [node for node in echoes_game_description.world_list.all_nodes if isinstance(node, LogbookNode)]

    player_pools = {
        0: create_player_pool(rng, default_layout_configuration, 0, 1),
    }
    initial_pickup_count = len(player_pools[0].pickups)

    patches = echoes_game_description.create_game_patches()
    patches = patches.assign_hint(
        logbook_nodes[0].resource(), Hint(HintType.LOCATION, None, PickupIndex(0))
    )
    action_log = (MagicMock(), MagicMock())
    player_state = MagicMock()
    player_state.index = 0
    player_state.game = player_pools[0].game
    player_state.pickups_left = runner._split_expansions(player_pools[0].pickups)[0]
    player_state.scan_asset_initial_pickups = {}

    mock_retcon_playthrough_filler.return_value = {player_state: patches}, action_log

    # Run
    filler_result = runner.run_filler(rng, player_pools, status_update)

    assert filler_result.action_log == action_log
    assert len(filler_result.player_results) == 1
    result_patches = filler_result.player_results[0].patches
    remaining_items = filler_result.player_results[0].unassigned_pickups

    # Assert
    assert len(result_patches.hints) == len(logbook_nodes)
    assert [hint for hint in result_patches.hints.values()
            if hint.precision is None] == []
    assert initial_pickup_count == len(remaining_items) + len(result_patches.pickup_assignment.values())


def test_fill_unassigned_hints_empty_assignment(echoes_game_description):
    # Setup
    rng = Random(5000)
    base_patches = echoes_game_description.create_game_patches()
    expected_logbooks = sum(1 for node in echoes_game_description.world_list.all_nodes
                            if isinstance(node, LogbookNode))

    # Run
    result = runner.fill_unassigned_hints(base_patches,
                                          echoes_game_description.world_list,
                                          rng, {})

    # Assert
    assert len(result.hints) == expected_logbooks


def test_add_hints_precision(empty_patches, mocker):
    failed_relative_provider = MagicMock(return_value=None)
    relative_hint_provider = MagicMock()
    mocker.patch("randovania.generator.filler.runner._get_relative_hint_providers",
                 return_value=[failed_relative_provider, relative_hint_provider])

    player_state = MagicMock()
    rng = MagicMock()
    hints = [
        Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.DETAILED,
                                              HintItemPrecision.DETAILED, include_owner=False), PickupIndex(1)),
        Hint(HintType.LOCATION, None, PickupIndex(2)),
        Hint(HintType.LOCATION, None, PickupIndex(3)),
    ]

    initial_patches = empty_patches
    for i, hint in enumerate(hints):
        initial_patches = initial_patches.assign_hint(LogbookAsset(i), hint)

    # Run
    result = runner.add_hints_precision(player_state, initial_patches, rng)

    # Assert
    failed_relative_provider.assert_called_once_with(player_state, initial_patches, rng, PickupIndex(2))
    relative_hint_provider.assert_called_once_with(player_state, initial_patches, rng, PickupIndex(3))
    assert result.hints == {
        LogbookAsset(0): Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.DETAILED,
                                                               HintItemPrecision.DETAILED, include_owner=False),
                              PickupIndex(1)),
        LogbookAsset(1): Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.WORLD_ONLY,
                                                               HintItemPrecision.PRECISE_CATEGORY, include_owner=False),
                              PickupIndex(2)),
        LogbookAsset(2): relative_hint_provider.return_value,
    }


def _make_pickup(item_category: ItemCategory):
    return PickupEntry(
        name="Pickup",
        model_index=0,
        item_category=item_category,
        broad_category=item_category,
        resources=(
            ConditionalResources(None, None, ()),
        ),
    )


@pytest.mark.parametrize("location_precision", [HintLocationPrecision.RELATIVE_TO_AREA,
                                                HintLocationPrecision.RELATIVE_TO_INDEX])
def test_add_relative_hint(echoes_game_description, empty_patches, location_precision):
    # Setup
    rng = Random(5000)
    target_precision = MagicMock()
    precise_distance = MagicMock()
    precision = MagicMock()
    patches = empty_patches.assign_pickup_assignment({
        PickupIndex(8): PickupTarget(_make_pickup(ItemCategory.MOVEMENT), 0),
    })

    if location_precision == HintLocationPrecision.RELATIVE_TO_AREA:
        max_distance = 8
        data = RelativeDataArea(
            precise_distance,
            AreaLocation(0x3BFA3EFF, 0x62AC8AC4),
            precision,
        )
    else:
        max_distance = 20
        data = RelativeDataItem(
            precise_distance,
            PickupIndex(8),
            precision,
        )

    # Run
    result = runner.add_relative_hint(echoes_game_description.world_list,
                                      patches,
                                      rng,
                                      PickupIndex(1),
                                      target_precision,
                                      location_precision,
                                      precise_distance,
                                      precision,
                                      max_distance=max_distance)

    # Assert
    pair = PrecisionPair(location_precision, target_precision, include_owner=False, relative=data)
    assert result == Hint(HintType.LOCATION, pair, PickupIndex(1))



import asyncio
import dataclasses
import struct
from asyncio import StreamReader, StreamWriter
from typing import List, Optional, Dict

from randovania.game_connection.backend_choice import GameBackendChoice
from randovania.game_connection.connection_backend import ConnectionBackend, MemoryOperation, MemoryOperationException
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_description.world import World


@dataclasses.dataclass(frozen=True)
class SocketHolder:
    reader: StreamReader
    writer: StreamWriter
    api_version: int
    max_input: int
    max_output: int
    max_addresses: int


class RequestBatch:
    def __init__(self):
        self.data = b""
        self.ops = []
        self.num_read_bytes = 0
        self.addresses = []

    def copy(self) -> "RequestBatch":
        new = RequestBatch()
        new.data = self.data
        new.ops = list(self.ops)
        new.num_read_bytes = self.num_read_bytes
        new.addresses.extend(self.addresses)
        return new

    def build_request_data(self):
        header = struct.pack(f">BBBB{len(self.addresses)}I", 0, len(self.ops), len(self.addresses), 1, *self.addresses)
        return header + self.data

    @property
    def input_bytes(self):
        return len(self.data) + 4 * len(self.addresses)

    @property
    def num_validator_bytes(self):
        return 1 + (len(self.ops) - 1) // 8 if self.ops else 0

    @property
    def output_bytes(self):
        return self.num_read_bytes + self.num_validator_bytes

    def is_compatible_with(self, holder: SocketHolder):
        return (len(self.addresses) < holder.max_addresses
                and self.output_bytes <= holder.max_output
                and self.input_bytes <= holder.max_input)

    def add_op(self, op: MemoryOperation):
        if op.address not in self.addresses:
            self.addresses.append(op.address)

        if op.read_byte_count is not None:
            self.num_read_bytes += op.read_byte_count

        op_byte = self.addresses.index(op.address)
        if op.read_byte_count is not None:
            op_byte = op_byte | 0x80
        if op.write_bytes is not None:
            op_byte = op_byte | 0x40
        if op.byte_count == 4:
            op_byte = op_byte | 0x20
        if op.offset is not None:
            op_byte = op_byte | 0x10

        self.data += struct.pack(">B", op_byte)
        if op.byte_count != 4:
            self.data += struct.pack(">B", op.byte_count)
        if op.offset is not None:
            self.data += struct.pack(">h", op.offset)
        if op.write_bytes is not None:
            self.data += op.write_bytes

        self.ops.append(op)


def _was_invalid_address(response: bytes, i: int) -> bool:
    try:
        return not response[i // 8] & (1 << (i % 8))
    except IndexError:
        raise MemoryOperationException("Server response too short for validator bytes")


class NintendontBackend(ConnectionBackend):
    _world: Optional[World] = None
    _port = 43673
    _socket: Optional[SocketHolder] = None

    def __init__(self, ip: str):
        super().__init__()
        self._ip = ip

    @property
    def ip(self):
        return self._ip

    @property
    def lock_identifier(self) -> Optional[str]:
        return None

    @property
    def backend_choice(self) -> GameBackendChoice:
        return GameBackendChoice.NINTENDONT

    # Game Backend Stuff
    async def _connect(self) -> bool:
        if self._socket is not None:
            return True

        try:
            self._socket_error = None
            self.logger.info(f"Connecting to {self._ip}:{self._port}.")
            reader, writer = await asyncio.open_connection(self._ip, self._port)

            # Send API details request
            self.logger.info(f"Connection open, requesting API details.")

            writer.write(struct.pack(f">BBBB", 1, 0, 0, 1))
            await asyncio.wait_for(writer.drain(), timeout=30)

            self.logger.debug(f"Waiting for API details response.")
            response = await asyncio.wait_for(reader.read(1024), timeout=15)
            api_version, max_input, max_output, max_addresses = struct.unpack_from(">IIII", response, 0)

            self.logger.info(f"Remote replied with API level {api_version}, connection successful.")
            self._socket = SocketHolder(reader, writer, api_version, max_input, max_output, max_addresses)
            return True

        except (OSError, asyncio.TimeoutError, struct.error) as e:
            self._socket = None
            self.logger.warning(f"Unable to connect to {self._ip}:{self._port} - ({type(e).__name__}) {e}")
            self._socket_error = e

    def _disconnect(self):
        socket = self._socket
        self._socket = None
        if socket is not None:
            socket.writer.close()

    def _prepare_requests_for(self, ops: List[MemoryOperation]) -> List[RequestBatch]:
        requests: List[RequestBatch] = []
        current_batch = RequestBatch()

        def _new_request():
            nonlocal current_batch
            requests.append(current_batch)
            current_batch = RequestBatch()

        processes_ops = []
        max_write_size = self._socket.max_input - 20
        for op in ops:
            if op.byte_count == 0:
                continue
            op.validate_byte_sizes()

            if op.read_byte_count is None and (op.write_bytes is not None
                                               and len(op.write_bytes) > max_write_size):
                self.logger.debug(f"Operation had {len(op.write_bytes)} bytes, "
                                  f"above the limit of {max_write_size}. Splitting.")
                for offset in range(0, len(op.write_bytes), max_write_size):
                    if op.offset is None:
                        address = op.address + offset
                        op_offset = None
                    else:
                        address = op.address
                        op_offset = op.offset + offset
                    processes_ops.append(MemoryOperation(
                        address=address,
                        offset=op_offset,
                        write_bytes=op.write_bytes[offset:min(offset + max_write_size, len(op.write_bytes))],
                    ))
            else:
                processes_ops.append(op)

        for op in processes_ops:
            experimental = current_batch.copy()
            experimental.add_op(op)

            if not experimental.is_compatible_with(self._socket):
                _new_request()

            current_batch.add_op(op)
            if not current_batch.is_compatible_with(self._socket):
                raise ValueError(f"Request {op} is not compatible with current server.")

        # Finish the last batch
        _new_request()

        return requests

    async def _send_requests_to_socket(self, requests: List[RequestBatch]) -> List[bytes]:
        all_responses = []
        try:
            for request in requests:
                data = request.build_request_data()
                self._socket.writer.write(data)
                await self._socket.writer.drain()
                if request.output_bytes > 0:
                    response = await asyncio.wait_for(self._socket.reader.read(1024), timeout=15)
                    all_responses.append(response)
                else:
                    all_responses.append(b"")

        except (OSError, asyncio.TimeoutError) as e:
            if isinstance(e, asyncio.TimeoutError):
                self.logger.warning(f"Timeout when reading response from {self._ip}")
                self._socket_error = MemoryOperationException(f"Timeout when reading response")
            else:
                self.logger.warning(f"Unable to send {len(requests)} request to {self._ip}:{self._port}: {e}")
                self._socket_error = MemoryOperationException(f"Unable to send {len(requests)} requests: {e}")
            self._disconnect()
            raise self._socket_error from e

        return all_responses

    async def _perform_memory_operations(self, ops: List[MemoryOperation]) -> Dict[MemoryOperation, bytes]:
        if self._socket is None:
            raise MemoryOperationException("Not connected")

        requests = self._prepare_requests_for(ops)
        all_responses = await self._send_requests_to_socket(requests)

        result = {}

        for request, response in zip(requests, all_responses):
            read_index = request.num_validator_bytes
            for i, op in enumerate(request.ops):
                if op.read_byte_count is None:
                    continue

                if _was_invalid_address(response, i):
                    raise MemoryOperationException("Operation tried to read an invalid address")

                split = response[read_index:read_index + op.read_byte_count]
                if len(split) != op.read_byte_count:
                    raise MemoryOperationException(f"Received {len(split)} bytes, expected {op.read_byte_count}")
                else:
                    assert op not in result
                    result[op] = split

                read_index += op.read_byte_count

        return result

    async def update(self, dt: float):
        if not self._enabled:
            return

        if not await self._connect():
            return

        if not await self._identify_game():
            return

        await self._interact_with_game(dt)

    @property
    def name(self) -> str:
        return "Nintendont"

    @property
    def current_status(self) -> GameConnectionStatus:
        if self._socket is None:
            return GameConnectionStatus.Disconnected

        if self.patches is None:
            return GameConnectionStatus.UnknownGame

        if self._world is None:
            return GameConnectionStatus.TitleScreen
        elif not self.checking_for_collected_index:
            return GameConnectionStatus.TrackerOnly
        else:
            return GameConnectionStatus.InGame



import asyncio
import json
import logging
from pathlib import Path
from typing import List, Set, Optional, AsyncContextManager, Tuple

import pid
from PySide2.QtCore import QObject
from asyncqt import asyncSlot

from randovania.bitpacking import bitpacking
from randovania.game_connection.game_connection import GameConnection
from randovania.game_description import data_reader
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.games.prime import default_data
from randovania.gui.lib.qt_network_client import QtNetworkClient
from randovania.network_common.pickup_serializer import BitPackPickupEntry


class BackendInUse(Exception):
    def __init__(self, lock_file: Path):
        self.lock_file = lock_file


class Data(AsyncContextManager):
    collected_locations: Set[int]
    uploaded_locations: Set[int]
    latest_message_displayed: int

    def __init__(self, path: Path):
        self._path = path
        self._lock = asyncio.Lock()

        try:
            with self._path.open() as f:
                data = json.load(f)
            self.collected_locations = set(data["collected_locations"])
            self.uploaded_locations = set(data["uploaded_locations"])
            self.latest_message_displayed = data["latest_message_displayed"]
        except FileNotFoundError:
            self.collected_locations = set()
            self.uploaded_locations = set()
            self.latest_message_displayed = 0

    async def __aenter__(self) -> "Data":
        await self._lock.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        new_data = json.dumps({
            "collected_locations": list(self.collected_locations),
            "uploaded_locations": list(self.uploaded_locations),
            "latest_message_displayed": self.latest_message_displayed,
        })
        self._path.write_text(new_data)
        self._lock.release()


class MultiworldClient(QObject):
    _data: Optional[Data] = None
    _received_pickups: List[Tuple[str, PickupEntry]]
    _notify_task: Optional[asyncio.Task] = None
    _pid: Optional[pid.PidFile] = None

    def __init__(self, network_client: QtNetworkClient, game_connection: GameConnection):
        super().__init__()
        self.logger = logging.getLogger(__name__)

        self.network_client = network_client
        self.game_connection = game_connection
        self._pickups_lock = asyncio.Lock()

        pid_name = game_connection.backend.lock_identifier
        if pid_name is not None:
            self._pid = pid.PidFile(pid_name)
            try:
                self._pid.create()
            except pid.PidFileError as e:
                raise BackendInUse(Path(self._pid.filename)) from e
            self.logger.debug(f"Creating pid file at {self._pid.filename}")

        self._game = data_reader.decode_data(default_data.decode_default_prime2())

    @property
    def is_active(self) -> bool:
        return self._data is not None

    async def start(self, persist_path: Path):
        self.logger.debug("start")

        if self._pid is not None and self._pid.fh is None:
            self._pid.create()

        self._data = Data(persist_path)
        self.game_connection.set_location_collected_listener(self.on_location_collected)
        self.network_client.GameUpdateNotification.connect(self.on_network_game_updated)

        await self.on_network_game_updated()

        self.start_notify_collect_locations_task()

    async def stop(self):
        self.logger.debug("stop")

        if self._notify_task is not None:
            self._notify_task.cancel()
            self._notify_task = None

        self.game_connection.set_location_collected_listener(None)
        try:
            self.network_client.GameUpdateNotification.disconnect(self.on_network_game_updated)
        except RuntimeError:
            pass

        async with self._pickups_lock:
            self.game_connection.set_permanent_pickups([])

        self._data = None
        if self._pid is not None:
            self._pid.close()

    def _decode_pickup(self, data: bytes) -> PickupEntry:
        decoder = bitpacking.BitPackDecoder(data)
        return BitPackPickupEntry.bit_pack_unpack(decoder, self._game.resource_database)

    async def _notify_collect_locations(self):
        while True:
            locations_to_upload = tuple(self._data.collected_locations - self._data.uploaded_locations)
            if not locations_to_upload:
                break

            try:
                await self.network_client.game_session_collect_locations(locations_to_upload)
            except Exception as e:
                self.logger.exception(f"Exception {type(e)} when attempting to "
                                      f"upload {len(locations_to_upload)} locations.")
                await asyncio.sleep(1)
                continue

            async with self._data as data:
                data.uploaded_locations.update(locations_to_upload)

    def start_notify_collect_locations_task(self):
        if self._notify_task is not None and not self._notify_task.done():
            return

        self._notify_task = asyncio.create_task(self._notify_collect_locations())

    async def on_location_collected(self, location: int):
        if location in self._data.collected_locations:
            self.logger.info(f"{location}, but location was already collected")
            return

        self.logger.info(f"{location}, a new location")
        async with self._data as data:
            data.collected_locations.add(location)

        self.start_notify_collect_locations_task()

    async def refresh_received_pickups(self):
        self.logger.debug(f"start")
        async with self._pickups_lock:
            result = await self.network_client.game_session_request_pickups()

            self.logger.info(f"received {len(result)} items")
            self._received_pickups = [
                (provider_name, self._decode_pickup(data))
                for provider_name, data in result
            ]

    @asyncSlot()
    async def on_network_game_updated(self):
        await self.refresh_received_pickups()

        async with self._pickups_lock:
            self.game_connection.set_permanent_pickups(self._received_pickups)



import dataclasses
import datetime
from typing import List, Optional

import aiohttp
from PySide2.QtWidgets import QPushButton, QDialogButtonBox, QDialog, QTableWidgetItem
from asyncqt import asyncSlot

from randovania.gui.generated.racetime_browser_dialog_ui import Ui_RacetimeBrowserDialog
from randovania.gui.lib import common_qt_lib, async_dialog
from randovania.gui.lib.qt_network_client import handle_network_errors
from randovania.layout.permalink import Permalink


@dataclasses.dataclass(frozen=True)
class RaceEntry:
    name: str
    status: str
    verbose_status: str
    entrants: int
    goal: str
    opened_at: datetime.datetime
    info: str

    @classmethod
    def from_json(cls, data) -> "RaceEntry":
        return RaceEntry(
            name=data["name"],
            status=data["status"]["value"],
            verbose_status=data["status"]["verbose_value"],
            entrants=data["entrants_count"],
            goal=data["goal"]["name"],
            opened_at=datetime.datetime.fromisoformat(data["opened_at"].replace("Z", "+00:00")),
            info=data["info"],
        )


_RACES_URL = "https://racetime.gg/mp2r/data"
_TEST_RESPONSE = {
    "name": "Metroid Prime 2: Echoes Randomizer",
    "short_name": "MP2R",
    "slug": "mp2r",
    "url": "/mp2r",
    "data_url": "/mp2r/data",
    "image": None,
    "info": "Download the Metroid Prime 2: Echoes randomizer <a href=\"https://github.com/randovania/randovania/releases\">here</a>.\r\n<p><p>\r\n<b><u>Category Rules</b></u>\r\n<p><p>\r\nTiming starts when confirming the game's difficulty on the main menu (or selecting a new file, if you are not able to select a difficulty). \r\n<p><p>\r\n<b>Beat the game</b>: Timing stops once you have activated a cutscene, elevator, portal, or teleport that will lead to the credits or Mission Final screen without any further game input.\r\n <p>\r\n<b>Collect all Sky Temple Keys</b>: Timing stops once you have at least one of each available Sky Temple Key (Sky Temple Key 1, 2, 3, etc.) currently collected. \r\n<p>\r\n<b><u>Miscellaneous</b></u>\r\n<p><p>\r\nThe only permitted emulator is Dolphin, running on version 5.0 or later. Running the ISO on console via Nintendont is allowed.\r\n<p>\r\nISO inspection/ROM inspection/memory inspection/software reverse-engineering that determines item locations or elevator destinations of the race seed prior to or during the race is <b>banned</b>.\r\n<p>\r\nTurbo functions, increasing the disk read speed, save states, fast forward and disabling fog are all <b>banned</b>. However, any of these functions may be allowed if all racers permit it, or a category rule permits it.",
    "streaming_required": False,
    "owners": [
        {
            "id": "q5rNGD3DYAo9blOy",
            "full_name": "gollop#9668",
            "name": "gollop",
            "discriminator": "9668",
            "url": "/user/q5rNGD3DYAo9blOy",
            "avatar": None,
            "pronouns": "they/them",
            "flair": "moderator",
            "twitch_name": "golloppp",
            "twitch_display_name": "golloppp",
            "twitch_channel": "https://www.twitch.tv/golloppp",
            "can_moderate": True
        }
    ],
    "moderators": [],
    "current_races": [
        {
            "name": "mp2r/proud-arena-4584",
            "status": {
                "value": "open",
                "verbose_value": "Open",
                "help_text": "Anyone may join this race"
            },
            "url": "/mp2r/proud-arena-4584",
            "data_url": "/mp2r/proud-arena-4584/data",
            "goal": {
                "name": "Beat the game",
                "custom": False
            },
            "info": "r6Ds6EUgBAGQBwkUhAcUAIEjxlUA7Q==  || Seed Hash: Morph Defiled Minigyro (I6LNUHZH)",
            "entrants_count": 5,
            "entrants_count_inactive": 0,
            "opened_at": "2020-10-03T18:46:17.865Z",
            "started_at": None,
            "time_limit": "P1DT00H00M00S"
        }
    ]
}


async def _query_server() -> dict:
    async with aiohttp.ClientSession() as session:
        async with session.get(_RACES_URL) as response:
            response.raise_for_status()
            return await response.json()


class RacetimeBrowserDialog(QDialog, Ui_RacetimeBrowserDialog):
    races: List[RaceEntry]
    permalink: Optional[Permalink] = None

    def __init__(self):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self.refresh_button = QPushButton("Refresh")
        self.button_box.addButton(self.refresh_button, QDialogButtonBox.ResetRole)
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
        self.button_box.button(QDialogButtonBox.Ok).setText("Import")

        self.button_box.accepted.connect(self.attempt_join)
        self.button_box.rejected.connect(self.reject)
        self.refresh_button.clicked.connect(self.refresh)

        self._status_checks = (
            (self.status_open_check, "open"),
            (self.status_invitational_check, "invitational"),
            (self.status_pending_check, "pending"),
            (self.status_inprogress_check, "in_progress"),
            (self.status_finished_check, "finished"),
            (self.status_cancelled_check, "cancelled"),
        )
        for check, _ in self._status_checks:
            check.stateChanged.connect(self.update_list)
        self.filter_name_edit.textEdited.connect(self.update_list)

        self.table_widget.itemSelectionChanged.connect(self.on_selection_changed)
        self.table_widget.itemDoubleClicked.connect(self.on_double_click)

    @asyncSlot()
    @handle_network_errors
    async def refresh(self) -> bool:
        self.refresh_button.setEnabled(False)
        try:
            try:
                raw_races = await _query_server()
            except aiohttp.ClientError as e:
                await async_dialog.warning(self, "Connection error",
                                           f"Unable to retrieve races from `{_RACES_URL}`: {e}")
                return False

            self.races = [RaceEntry.from_json(item) for item in raw_races["current_races"]]
            self.update_list()
            return True
        finally:
            self.refresh_button.setEnabled(True)

    def on_selection_changed(self):
        self.button_box.button(QDialogButtonBox.Ok).setEnabled(len(self.table_widget.selectedItems()) > 0)

    @property
    def selected_race(self):
        row: int = self.table_widget.selectedIndexes()[0].row()
        return self.races[row]

    @asyncSlot(QTableWidgetItem)
    async def on_double_click(self, item: QTableWidgetItem):
        await self.attempt_join()

    @asyncSlot()
    @handle_network_errors
    async def attempt_join(self):
        if not self.races:
            return

        race = self.selected_race

        permalink = None

        for word in race.info.split(" "):
            try:
                permalink = Permalink.from_str(word)
            except ValueError:
                continue

        if permalink is None:
            await async_dialog.warning(self, "Missing permalink",
                                       "Unable to get a valid Permalink from this race's info.")
        else:
            self.permalink = permalink
            return self.accept()

    def update_list(self):
        self.table_widget.clear()
        self.table_widget.setHorizontalHeaderLabels(["Name", "Status", "Entrants",
                                                     "Goal", "Info", "Opened At"])

        name_filter = self.filter_name_edit.text().strip()

        displayed_status = set()
        for (check, status) in self._status_checks:
            if check.isChecked():
                displayed_status.add(status)

        visible_races = [
            race
            for race in self.races
            if (race.status in displayed_status
                and name_filter in race.name)
        ]

        self.table_widget.setRowCount(len(visible_races))
        for i, session in enumerate(visible_races):
            name = QTableWidgetItem(session.name)
            status = QTableWidgetItem(session.verbose_status)
            entrants = QTableWidgetItem(str(session.entrants))
            goal = QTableWidgetItem(session.goal)
            info = QTableWidgetItem(session.info)
            opened_at = QTableWidgetItem(session.opened_at.astimezone().strftime("%c"))

            self.table_widget.setItem(i, 0, name)
            self.table_widget.setItem(i, 1, status)
            self.table_widget.setItem(i, 2, entrants)
            self.table_widget.setItem(i, 3, goal)
            self.table_widget.setItem(i, 4, info)
            self.table_widget.setItem(i, 5, opened_at)

        self.status_label.setText(f"{len(self.races)} races total, {len(visible_races)} displayed.")
